function [bbs,masks,leafind] = fv_candbox_superpixelsinside( p, inpbbs, trees )
% [bbs,masks,leafind] = fv_candbox_superpixelsinside( p, inpbbs, trees )
%
% Using the baselevel superpixel segmentation from the provided segmentation trees,
% return segmentations for the provided boxes using the superpixels lying inside.
%
% INPUT
% p         
%   .ovth               Minimum "ratio of segmentation area inside the box" threshold. A much faster method is used for ovth==1.
%   .superpixelseg      each | avg
%   .ifempty            skip | zeros
% inpbbs                (ninpbb 4) Input bounding boxes.
% trees                 
%
% OUTPUT 
% -- each --
% bbs                   Cell array (ninpbb ntree)
% mask                  Cell array (ninpbb ntree)
%                       masks{i,j} contains binary mask wrt j-th tree for the i-th box. 
%                       If there are no segments inside and ifempty==skip, both bbs{i,j} and masks{i,j} are empty.
%                       If there are no segments inside and ifempty==zeros, both masks{i,j} = false(bbheight,bbwidth)
%                       bbs{i,j} contains either [] or just a copy of bbs(i,:). 
% [leafind]             leafind{i,j} contains a row vector leaf indices contributing to the i-th mask from the j-th tree.
% 
% -- avg --
% bbs                   Cell array (ninpbb 1)
% mask                  Cell array (ninpbb 1)
%                       masks{i} contains weight mask averaged over trees that provide non-empty masks. Weights are given in 'single'.
%                       If there are no segments inside and ifempty==skip, both bbs{i} and masks{i} are empty.
%                       If there are no segments inside and ifempty==zeros, both masks{i} = false(bbheight,bbwidth)
%                       That is, bbs{i} contains either [] or just a copy of bbs(i,:). 
% [leafind]             leafind{i,j} contains a col vector leaf indices contributing to the i-th mask from the j-th tree.
%
% R.Gokberk Cinbis 2012

if p.ovth~=1
    % find the area of each superpixel
    for ti = 1:length(trees)
        trees(ti).LeafArea = accumarray( trees(ti).leafmap(:), 1, [trees(ti).nleaf 1] );
    end
end

bbs = cell(size(inpbbs,1),length(trees));
masks = cell(size(inpbbs,1),length(trees));
leafind = cell(size(inpbbs,1),length(trees));

for bbi = 1:size(inpbbs,1)
    bb = inpbbs(bbi,:);

    for ti = 1:length(trees)
        ssbbs = vertcat(trees(ti).bbs(1:trees(ti).nleaf,:)); % superpixel bounding boxes.

        maskbb = []; % important!

        if p.ovth==1

            m = ssbbs(:,1)>=bb(1) & ssbbs(:,2)>=bb(2) & ssbbs(:,3)<=bb(3) & ssbbs(:,4)<=bb(4);
            maskbb = [min(ssbbs(m,[1 2]),[],1) max(ssbbs(m,[3 4]),[],1)];

            if ~isempty(maskbb)
                cur_leaves = colvec( find(m) );
                mask = ismember( trees(ti).leafmap( maskbb(2):maskbb(4), maskbb(1):maskbb(3) ) , cur_leaves );
                leafind{bbi,ti} = cur_leaves;
            end

        else

            % find the area within the box of each superpixel
            subarea = accumarray( colvec( trees(ti).leafmap( bb(2):bb(4), bb(1):bb(3) ) ), 1, [trees(ti).nleaf 1] );

            % now check the ones that a mostly inside
            cur_leaves = colvec( find( (subarea ./ trees(ti).LeafArea) >= p.ovth ) );

            if ~isempty(cur_leaves)
                mask = ismember( trees(ti).leafmap, cur_leaves );
                [y,x] = find(mask);
                maskbb = [min(x) min(y) max(x) max(y)];
                mask = mask( maskbb(2):maskbb(4), maskbb(1):maskbb(3) );
                leafind{bbi,ti} = cur_leaves;
            end

        end

        if ~isempty( maskbb ) % has some foreground pixels inside
            mask = fv_candbox_cropandpadmask( bb, mask, maskbb );

            bbs{bbi,ti} = bb;
            masks{bbi,ti} = mask;
        end

    end
end

m = cellfun(@isempty,bbs);
assert( isequal(m,cellfun(@isempty,masks)) );

switch(p.superpixelseg)
    case 'each'
        % nothing to do.
        is_bool = true;

    case 'avg'
        ind = find( any(~m,2) );
        is_bool = false;

        bbs2 = cell(size(inpbbs,1),1);
        masks2 = cell(size(inpbbs,1),1);
        for bbi = rowvec(ind)
            mx = ~m(bbi,:);
            assert(any(mx));
            masks2{bbi} = single( mean(cat( 3, masks{bbi,mx} ),3) ); % store weights in single after taking mean()
            bbs2{bbi} = inpbbs(bbi,:);
        end

        bbs = detach(bbs2);
        masks = detach(masks2);

    otherwise
        error('unknown');
end

switch(p.ifempty)
    case 'skip'
        % nothing to do

    case 'zeros'
        [ind_i,ind_j] = find( cellfun(@isempty,bbs) );
        for k = 1:length(ind_i)
            i = ind_i(k);
            j = ind_j(k);

            bb = inpbbs(i,:);
            assert(isempty(bbs{i,j}));
            assert(isempty(masks{i,j}));
            bbs{i,j} = bb;

            if is_bool
                masks{i,j} = false( bb(4)-bb(2)+1, bb(3)-bb(1)+1 );
            else
                masks{i,j} = zeros( bb(4)-bb(2)+1, bb(3)-bb(1)+1, 'single' );
            end

        end

    otherwise
        error('unknown p.ifempty');
end

assert(length(masks)==size(bbs,1));

function m2 = fv_candbox_cropandpadmask( bb, mask, maskbb )
% remove parts of the mask that are not inside the box.
% if mask is smaller that the box, pad it with zeros.
%
% R.Gokberk Cinbis, 2012

bb = double(bb);
maskbb = double(maskbb);

assert( size(mask,1) == (maskbb(4)-maskbb(2)+1) );
assert( size(mask,2) == (maskbb(3)-maskbb(1)+1) );

padbeg = max(maskbb([2 1]) - bb([2 1]),0); % hw
padend = max(bb([4 3]) - maskbb([4 3]),0); % hw

a = [max(bb([1 2]),maskbb([1 2])) min(bb([3 4]),maskbb([3 4]))] - maskbb([1 2 1 2]) + 1;
m = mask( a(2):a(4), a(1):a(3) );
m2 = false( size(m) + padend + padbeg );
m2( (padbeg(1)+1):(padbeg(1)+size(m,1)), (padbeg(2)+1):(padbeg(2)+size(m,2)) ) = m;

assert( size(m2,1) == (bb(4)-bb(2)+1) );
assert( size(m2,2) == (bb(3)-bb(1)+1) );
