function [LogL,MoFA,Q] = mfa(X,C,ops,MoFA)
%
% [LogL,mfa,Q] = mfa(X,d,C,iso,dia);
%
% X     : data matrix (N data points in D dimensions as columns)
% C     : number of mixture components
% ops   : options for learning algorithm
% MoFA  : initialize from given Mixture of Factor Analyzers
%
% LogL  : row vector of log-likelihoods after each EM iteration
% MoFA  : structure containing the Mixture of Factor Analyzers
%         MoFA.W (D x d x C) : factor loading matrices
%         MoFA.M (D x C)     : component means
%         MoFA.Psi ( D x C ) : component noise variances
%         MoFA.mix ( C x 1 ) : mixing weights
% Q     : C x N matrix of component posterior probabilities
%
% by Jakob Verbeek, INRIA, 2005-2011.
%
% TODO: 
% - check-out NaN logl cases.

%{
addpath /tmp/cinbis/visible/projects/fishervec/kmeans
cf; N=20; D=5; cn_setrandseed(100); [L,M] = mfa(rand(D,N),3,struct2('dim',3,'iso',1,'eqn',0,'dia',1,'minvar',eps,'maxbyte',inf));
%}


ops.C = C;
[D,N] = size(X);

ops = mfa_opsdefaults(ops);

if ops.dia
  fprintf('--> running EM for MoFA\n');
  fprintf('    mixture components: %d\n',C);
  fprintf('    latent dimensions : %d\n',ops.dim);
  fprintf('    data: %d points in %d dimensions\n',N,D);
  if ops.iso; fprintf('    using isotropic noise model (ie Probabilistic PCA)\n');end
  if ops.eqn; fprintf('    equal noise model in components\n');end
end

if ~isinf(ops.sparsityk); 
    if ops.sparsityk<1; error('sparsityk<1'); end
    if ops.sparsityk>C; error('sparsityk>C'); end
    mywarn(sprintf('posteriors will be approximated with top %d elements\n',ops.sparsityk));
end

if nargin < 4;    MoFA = init_mfa(X,C,ops);  end % Initialisation

X2   = X.^2;
LogL = zeros(ops.mxi,1);
iter = 0;
done = false; 

XT = [];
if isinf(ops.maxbyte) 
    splitind = [];

    if (ops.dim>0) && ops.iso
        XT = X.';
    end
else
    splitind = cn_splitindices(N, min(ceil(C*N*8*2/ops.maxbyte),N) );
    fprintf('mfa -- #split=%d\n',length(splitind));
end

while ~done  
  iter = iter +1;  

  % make sure no dependent parameters leaked
  assert(isempty(intersect(fieldnames(MoFA),{'psiI','psiIM','sum_psiIM2','latprep'})));

  Q = []; % save memory
  if isempty(splitind)
      [Q, LogL(iter),Lats, lat_covs] = mfa_E_step(X,MoFA,X2,ops);
      MoFA                           = mfa_M_step(X,X2,detach(Q),Lats,lat_covs,ops,XT);  % M-step 
  else
      debug = 0;

      if debug
          [xQ, xLogL, xLats, xlat_covs] = mfa_E_step(X,MoFA,X2,ops);
          xMoFA                         = mfa_M_step(X,X2,(xQ),xLats,xlat_covs,ops,XT);  % M-step 
      end

      [LogL(iter),MoFA] = mfa_EM_chunked(ops,X,MoFA,X2,splitind);

      if debug
          if abs(xLogL-LogL(iter)) > 1e-11; keyboard; else; disp('chunk LogL: OK'); end
          if any(abs(MoFA.M(:)-xMoFA.M(:)) > 1e-12); keyboard; else; disp('chunk M: OK'); end
          if any(abs(MoFA.Psi(:)-xMoFA.Psi(:)) > 1e-12); keyboard; else; disp('chunk Psi: OK'); end
          if any(abs(MoFA.W(:)-xMoFA.W(:)) > 1e-12); keyboard; else; disp('chunk Psi: OK'); end
      end

  end

  done  = check_state(LogL(1:iter),ops,[]);       % check for convergence, plot diagnostics

  % make sure no dependent parameters leaked
  assert(isempty(intersect(fieldnames(MoFA),{'psiI','psiIM','sum_psiIM2','latprep'})));

end

LogL = LogL(1:iter);

