function [A,M,E] = kmeans(X,K,iters,dia,maxbyte)
% kmeans clustering
%
% [A,M,E] = kmeans(X,k,iters)
%
% X     - (d x n) d-dimensional input data
% K     -  number of centers, or (dxK) matrix for assignment to given means
% iters - number of iterations
% [dia] - give 1 to print info
% [maxbyte] - default=inf. maximum memory usage in helper(). ex: 4e9 for ~4GB.
%
% returns
% E  - sum of squared distances to nearest mean
% M  - (k x d) matrix of cluster centers
% A  - (n x 1) index of nearest center for each data point
%
% Jakob Verbeek, 2009-2011
%
% Gokberk: added maxbyte option.

N       = size(X,2);
iter    = 1;
E       = zeros(iters,1);

if numel(K) > 1;     % initialize with given means, compute assignment only
  M = K;
  K = size(M,2);  
  done    = 1;
else
  M = X(:,randi(N,K,1));
  done    = 0;
end

if nargin<4;dia=0;end
if nargin<5;maxbyte=inf;end

if isinf(maxbyte)
    splitind = [];
else
    splitind = cn_splitindices(N, min(ceil(K*N*8/maxbyte),N) );
    fprintf('kmeans -- #split=%d\n',length(splitind));
end

SumNormsX = sum(X(:).^2);

[Ei A] = helper(K,N,splitind,M,X);

E(iter) =-2*sum(Ei) + SumNormsX;

while ~done
  iter = iter +1; % increase update iteration nr.
  Aold = A;       % store old assignments
  
  % update centers
  C = accumarray(A,1,[K 1])'; % counts of the centers
  
  M = X*sparse(1:N,A,1,N,K);  % sum associated data points
  M = bsxfun(@times,M,C.^-1); % divide by counts
  
  Cempty = find(C==0);        % identify empty clusters
  M(:,Cempty) = X(:,randi(N,numel(Cempty),1)); % initialize them at random data points
  
  % update assignments
  [Ei A] = helper(K,N,splitind,M,X);

  if 0
      % seems OK.
      assert(~isempty(splitind)); [xEi xA] = helper(K,N,[],M,X); 
      xerr1 = sum(xA~=A); xerr2 = max(abs(Ei(:)-xEi(:)));
      if xerr1>0 || xerr2>1e-10; xerr1, xerr2, [A(:) xA(:) Ei(:) xEi(:)]', pause2; else; disp('OK'); end
  end
  
  % update error
  E(iter) =-2*sum(Ei) + SumNormsX;
  
  if iter >= iters;              done=1; disp('maxiter reached'); end % terminate after given nr of iterations
  if (iter>1) && (all(Aold==A)); done=1; disp('assignments didnt change'); end % terminate when no change in assignments  
  if ~isempty(Cempty);           done=0;end % force one more update if empty clusters have been re-initialized
  
  if (dia); fprintf('Iter %d, error %f\n',iter, E(iter));    end
  if cn_timepassth; fprintf('Iter %d, error %f\n',iter, E(iter)); end
  
end
E = E(1:iter);

fprintf('Kmeans finished with %d iters, error %f\n',iter, E(iter)); 

% sort centers by frequency
if numel(K)==1 
  
  C = accumarray(A,1,[K 1]); % counts of the centers
  [~, ii] = sort(C,'descend');
  M = M(:,ii);
  A2 = zeros(size(A));
  for i=1:K
    A2(A==ii(i)) = i;
  end
  A=A2;
end




function [Ei A] = helper(K,N,splitind,M,X)
% gokberk: added chunking. assumes N>>K (otherwise it will still work)

NormsM = sum(M.^2,1)/2; % (D K)

% checked: maxbyte=1e3; K=100;N=250;D=80;M=rand(D,K);X=rand(D,N);
% both codes seems equivalent.
%
% example: maxbyte=4e9; K=2048; N=1.5e6; D=80; --> 7 splits.

if isempty(splitind)
    T = bsxfun(@plus, M'*X, -NormsM'); % (K N) -- large!
    [Ei A] = max(T,[],1);
else
    % OK equivalent.
    for i = length(splitind):-1:1
        jj = splitind{i};
        T = bsxfun(@plus, M'*X(:,jj), -NormsM'); % (K N) -- large!
        [Ei(1,jj) A(1,jj)] = max(T,[],1);
    end
end
A=A';    

