function classifier = cn_pqliblinear_train(pqcenters,z2c,codes,wscr,extras,labels,p)
% classifier = cn_pqliblinear_train(pqcenters,codes,wscr,extras,labels,p)
%
% Liblinear based training on the top of PQ codes. Allows adding uncompressed dimensions and setting per-example
% weights.  Allows multiple code blocks to re-use the same PQ centers (Z dimension).
%
% Allows pqcenters to be provided in blocks with varying ds and nsq. This allows us to:
% a) Respect original data blocks with varying subquantizer dimensionality.
% b) Combine pq-compressed features of varying dimensionalities.
%
% PQ centers can be re-used using z2c to avoid multiple copies of the same pq-centers array.
% This reduces memory-CPU communication.
%
% Codes are stored in concatenated format for efficiency. 
% Then, codes has size (nsq_total N) which is vertcat((nsq_1 N),...,(nsq_Z N)) (mapped using z2c),
% and nsq_total = sum(nsq_c(z2c)).  First nsq_1 codes refer to pqcenters{z2c(1)}, following nsq_2 codes refer to
% pqcenters{z2c(2)} and so on.
%
% INPUT
% pqcenters     {(ds_c ks nsq_c)}       single  PQ centers as a (1 C) cell array. 
% z2c           (1 Z)                   double  Map codes to pqcenters. Using this, we can re-use pqcenters.
% codes         (nsq_total N)           uint8   PQ codes. nsq_total=sum(nsq_c(z2c))
% wscr          (Z N)|(1 N)|(Z 1)|(1 1) double  Weights to scale PQ compressed descriptors. 
%                                               Therefore, desc_z = decode(centers,codes_z) * wscr_z.
% extras        (de N)                  double  Uncompressed additional descriptors. Set to ones(1,N) for a standard bias term.
% labels        (1 N)                   double  Training labels. One-vs-all is used for multiclass.
% p
%   .samplew            Per-sample cost weights. Will NOT be normalized. 
%   .[costp]            Global cost parameter. Equivalent to setting samplew=samplew*costp. (def=1)
%   .clstype            1 (l2-reg,hinge^2-loss) | 3 (l2-reg,hinge-loss)
%   .[epsilon]          Convergence threshold. (def=0.1 for 2 and 3).
%   .[quiet]            Quiet training mode.
%
% OUTPUT
% classifier
%   .alls_pq            (ncls ks nsq_total)         Concatenated scores array.
%                                                   alls_pq = cat(3,...(nsq ks nsq_i)...)
%   .allw_e             (de ncls)                   Weights for "extras".
%   .allw_pq            (sum(ds_i.*nsq_i) ncls)     Per-subquantizer weight. Each column allw_pq(:,c) is 
%                                                   vertcat( ...,colvec( (ds_i nsq_i) ),... )
%                                                   This array can be discarded in most cases.
%   .idx_to_orglabel    (1 ncls)
%   .nsq_z              (1 Z),uint64                Defines the size of each code block.
%
% NOTES
% * For binary training, output has weights for a single class. Second class would have "-w" for all weights.
%
% R.Gokberk Cinbis, May 2012

% Bug Fix 2012-09-24: Now randomization within mex function is stable, it depends on #training samples. Previously, training results were slightly changing at each run.

% todo:
%   .[norminfo]         (def=[]) Linear normalization info, as returned by cn_normalizedata()
%                       Or parameters. Ex: {{'range',[0 1]}} or {'range01'}
%                       *** This will be embedded into classifier.all(w,b) ***
%                       *** Assumes that normalization has NOT been applied to TRAIN/VAL/TEST data. ***
%
% todo:
%   Would it affect the results significantly had we use codes:(Z nsq N) instead? The advantage is that we 
%   process a subquantizer across all Z, reducing memory-cache transfers of pq centers.

tt = tic;

%%%% check input 

% make sure unsupported fields are not passed
unsupported = {'equalw','initequalclstotalw','biasterm','ignoresampleth','shape','regtype','losstype','modelselection'};
for i = 1:length(unsupported)
    f = unsupported{i};
    assert(~isfield(p,f),['unsupported field: p.' f]);
end
assert(~isfield(p,'norminfo'),'norminfo not supported yet');

p = cn_setfielddefaults(p,true,'costp',1,'epsilon',[],'quiet',false);
% todo: quiet

% default convergence threshold
if isempty(p.epsilon)
    switch(p.clstype)
        case {1,3}
            p.epsilon = 0.1;
        otherwise
            error('unsupported p.clstype');
    end
end

% dimensions

assert(iscell(pqcenters));
C = length(pqcenters);
for fi = C:-1:1
    [ds_c(fi),ks,nsq_c(fi)] = size(pqcenters{fi}); 
    assert(ks==256);
end
% nsq_total_centers = sum(nsq_c);

% mapping
Z = length(z2c);
ds_z = ds_c(z2c);
nsq_z = nsq_c(z2c);
nsq_total = sum(nsq_z);

[nsq2,N] = size(codes);
assert(nsq_total==detach(nsq2));

[wscr_d1,wscr_d2,wscr_d3] = size(wscr);
assert((wscr_d1==Z) | (wscr_d1==1),'wscr can only have one of the following sizes: (Z 1),(Z N),(1 N),(1 1)');
assert((wscr_d2==N) | (wscr_d2==1),'wscr can only have one of the following sizes: (Z 1),(Z N),(1 N),(1 1)');
assert(wscr_d3==1,'wscr can only have one of the following sizes: (Z 1),(Z N),(1 N),(1 1)');
[de,N2] = size(extras);
assert(N==detach(N2),'Inconsistency in codes and extras size. Make sure codes has size (nsq_total N)');
labels = reshape(labels,1,N);

% sample samplew with cost parameter
p.samplew = reshape(p.samplew,1,N) * p.costp;

% multiclass training
idx_to_orglabel = rowvec(unique(labels));
idx_to_orglabel = [idx_to_orglabel(idx_to_orglabel>0) idx_to_orglabel(idx_to_orglabel<=0)];
ncls = length(idx_to_orglabel);

if ncls == 2
    ncls2 = 1;
else
    ncls2 = ncls;
end

% init output
allw_pq = zeros( sum(ds_z.*nsq_z), ncls2 );
allw_e = zeros(de,ncls2);

for clsj = 1:ncls2
    y = sign( double(labels == idx_to_orglabel(clsj))-.1 );
    [allw_pq(:,clsj),allw_e(:,clsj)] = cn_pqliblinear_mex(pqcenters,z2c,codes,wscr,extras,y,p.samplew,p.clstype,p.epsilon,p.quiet);
end

classifier = [];
classifier.allw_pq = allw_pq; 
classifier.allw_e = allw_e;
classifier.idx_to_orglabel = idx_to_orglabel;
classifier.nsq_z = uint64(nsq_z);

% convenient to have these here.
classifier.ds_c = ds_c;
classifier.nsq_c = nsq_c;
classifier.z2c = z2c;

%fprintf('cn_pqliblinear_train: train(): '); toc(tt); % all time until cn_pqliblinear_prep()
%tt = tic;

classifier = cn_pqliblinear_prep(classifier,pqcenters);

%fprintf('cn_pqliblinear_train: prep(): '); toc(tt);

