/**
 * alls = cn_pqliblinear_prepmex(allwt,z2c,centers)
 *
 * INPUT
 * allwt     (ncls sum(ds_i.*nsq_i)). (T)
 * z2c       (1 Z), base 1, uint64.
 * centers   {(ds_c ks nsq_c)}. (T)
 *
 * OUTPUT
 * alls      (ncls ks nsq_total). (T)
 *
 */

/**
 * Author: R. Gokberk Cinbis.
 */


#include "mex.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <cstring>
#include <stdint.h> // to get uint8_t
#include <limits>
#include <cstdarg>
#include "cn_matlabarrays.hpp"
#include "cn_macros.hpp"
#include "cn_threads.hpp"
#include "cn_pq.hpp"

using std::runtime_error;
using std::vector;
using cn::mex::getcell;
using cn::mex::numel;
using cn::mex::ndims;
using cn::mex::getscalar;
const uint64_t ks = cn::pq::ks;

/**
 * Matlab arguments.
 */
template <class T>
struct matargs_t {
    uint64_t ncls;

    uint64_t C;
    vector <uint64_t> ds_c;  // (1 C)
    vector <uint64_t> nsq_c; // (1 C)

    uint64_t Z;
    vector <uint64_t> z2c; // (1 Z) base 0.

    uint64_t nsq_total;
    uint64_t d_total;

    const T *allwt; // (ncls d_total)
    vector <const T *> centers; // (1 C)

    matargs_t() {}; // default constructor
    matargs_t( const matargs_t &from ); // disable copy constructor
};


// requires: args.C, d_total
// se
template<class T>
void getz2c(const mxArray *m_z2c,matargs_t <T> &args) {
    args.Z = numel(m_z2c);
    args.z2c.reserve(args.Z);
    args.nsq_total = 0;
    uint64_t d_total = 0;
    const uint64_t *z2c = cn::mex::getDenseData<uint64_t>( m_z2c );
    for ( uint64_t z = 0 ; z < args.Z; ++z ) {
        uint64_t c = z2c[z];
        if ( c > args.C )
            mexErrMsgTxt("invalid z2c entry, has to be <= C");
        if ( c < 1 )
            mexErrMsgTxt("invalid z2c entry, has to be base 1");
        c -= 1; // convert to base 0.
        args.z2c.push_back( c ); 
        args.nsq_total += args.nsq_c[c];
        d_total += args.ds_c[c]  * args.nsq_c[c];
    }
    cn::mex::assert2( (d_total == args.d_total), "Failed: d_total==sum(ds_c(z2c))" );
}


template<class T>
void getcenters(const mxArray *m_centers,matargs_t<T> &args) {
    args.C = numel(m_centers);
    args.ds_c.reserve(args.C);
    args.nsq_c.reserve(args.C);
    args.centers.reserve(args.C);
    for ( uint64_t c = 0; c < args.C; ++c ) {
        const mxArray *A = getcell(m_centers,c);
        args.ds_c.push_back( cn::mex::getdim(A,1) );
        args.nsq_c.push_back( cn::mex::getdim(A,3) );
        cn::mex::assert2( ndims(A) <= 3, "ndims(centers{c}) <= 3" );
        cn::mex::assert2( cn::mex::getdim(A,2) == ks, "ks incompability in centers array" );
        args.centers.push_back( cn::mex::getDenseData<T>(A) );
    }
}

template <class T>
mxArray *process(const mxArray *m_allwt,const mxArray *m_z2c,const mxArray *m_centers) {
    matargs_t <T> args;

    // m_allwt
    args.ncls = cn::mex::getdim(m_allwt,1);
    args.d_total = cn::mex::getdim(m_allwt,2);
    cn::mex::assert2(ndims(m_allwt) == 2);
    args.allwt = cn::mex::getDenseData<T>(m_allwt);

    // centers & z2c
    getcenters(m_centers,args);
    getz2c(m_z2c,args);

    // alloc output
    mxArray * m_out = cn::mex::createArrayZeros(cn::mex::type2MxClassID<T>(),args.ncls,ks,args.nsq_total);
    T *alls = cn::mex::getDenseData<T>(m_out);
    cn::pq::linclslookup(
            args.ds_c,
            args.nsq_c,
            args.centers,
            args.z2c,
            args.ncls,
            args.allwt,
            alls);

    return m_out;
}

enum { a_allwt = 0, a_z2c, a_centers };
void mexFunction(int nlhs, mxArray *plhs[],int nrhs, const mxArray *prhs[])
{
    if ( nrhs != 3 )
        mexErrMsgTxt("invalid number of input");
    if ( nlhs != 1 )
        mexErrMsgTxt("invalid number of output");

    const mxArray *m_allwt = prhs[a_allwt];
    const mxArray *m_z2c = prhs[a_z2c];
    const mxArray *m_centers = prhs[a_centers];

    try {

        // whether the centers are in double or in single.
        if ( numel(m_centers) == 0 )
            mexErrMsgTxt("centers array cannot be empty");
        mxClassID T = mxGetClassID( getcell(m_centers,0) );

        if ( T == mxDOUBLE_CLASS ) {
            plhs[0] = process<double>(m_allwt,m_z2c,m_centers);
        } else { 
            plhs[0] = process<float>(m_allwt,m_z2c,m_centers);
        }

    } catch(runtime_error &err) { 
        mexErrMsgTxt(err.what());
    }

}


