
/**
 * Author: R. Gokberk Cinbis.
 */


#include "mex.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <cstring>
#include <omp.h>
#include "cn_threads.hpp"
#include <stdint.h> // to get uint8_t
#include <limits>

// mxUint8 is not defined in mex.h. So, we use uint8_t

// compile with cn_makeall pq_assign

void process(const mwSize ds, const mwSize ks, const mwSize m, const mwSize N, const float *X, const float *cent3s, uint8_t *out ) {

    cn::setmaxthreads();

    const float float_inf = std::numeric_limits<float>::infinity();
    const mwSize fullD = ds*m;

#pragma omp parallel for 
    for (mwIndex xi=0; xi < N; ++xi) {  // over vectors
        const float *x0 = X + fullD*xi;
        const float *cent = cent3s;
        uint8_t *curout = out + m*xi;
        for (mwIndex sqi=0; sqi < m; ++sqi) { // over subvectors in the current x vector

            uint8_t nearestk = 0; // actually, no need to initialize but to avoid compiler warning..
            float nearestdist = float_inf;
            for (mwIndex k=0; k < ks; ++k) { // over clusters in the current subquantizer

                const float *x = x0;
                float tmp = 0;
                for (mwIndex d=0; d < ds; ++d ) { // over dimensions 
                    float diff = (*x - *cent);
                    tmp += diff * diff;
                    ++x;
                    ++cent;
                }

                // check if this is the new nearest cluster
                if ( tmp < nearestdist) {
                    nearestk = k;
                    nearestdist = tmp;
                }
                
                // debug: printf("sqi=%d k=%d dist=%.5f\n",sqi,k,tmp);

                // cent is now pointing to the next cluster
            }

            // debug: printf("sqi=%d nearestk=%d dist=%.5f\n",sqi,nearestk,nearestdist);

            *curout = nearestk;

            // cent is now pointing to the first cluster in the next subquantizer

            // next subvector
            x0 += ds;
            ++curout;
        }
    }
}



#define iPQ 0
#define iX 1

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{

    /*  check for proper number of arguments */
    if (nrhs!=2) {
        mexErrMsgTxt("There should be 2 input arguments.");
    } 
    if(nlhs!=1) {
        mexErrMsgTxt("There should be 1 output argument.");
    }

    /*  check for input type */
    if ( ! mxIsStruct( prhs[iPQ] ) ) {
        mexErrMsgTxt("First argument should be PQ struct");
    }

    if (mxGetClassID(prhs[iX]) != mxSINGLE_CLASS) {
        mexErrMsgTxt("Second argument should be single");
    }

    if (mxIsSparse(prhs[iX])) { // in case matlab adds supports for sparse single 
        mexErrMsgTxt("Sparse data is not accepted");
    }

    // get X data 
    mwSize D = mxGetM(prhs[iX]);
    mwSize N = mxGetN(prhs[iX]);
    const float *X = (float *) mxGetData(prhs[iX]);

    // get PQ data
    const mxArray *cent3sArr = mxGetField(prhs[iPQ],0,"cent3s");
    if (cent3sArr==NULL) {
        mexErrMsgTxt("PQ argument is invalid. Cannot acces cent3s field. Be sure to prepare it with pq_prep()");
    }
    if (mxGetClassID(cent3sArr) != mxSINGLE_CLASS) {
        mexErrMsgTxt("cent3sArr should be a float");
    }
    const float *cent3s = (const float *) mxGetData(cent3sArr);
    const mwSize *c3sz = mxGetDimensions(cent3sArr);
    mwSize ds = c3sz[0]; // number of elements in each subvector
    mwSize ks = c3sz[1]; // number of clusters in each subquantizer
    mwSize m;
    if (mxGetNumberOfDimensions(cent3sArr)<3) {
        m = 1;
    } else {
        m = c3sz[2]; // number of subquantizers
    }

    // always 256 clusters per subquantizer
    if ( ks != 256 ) {
        mexErrMsgTxt("In the current implementation ks is fixed to 256 and output is fixed to uint8. ks is incompatible.");
    }

    if ( !( std::numeric_limits<float>::has_infinity ) ) {
        mexErrMsgTxt("Needs infinity");
    }

    // sanity check
    if ( D != (ds*m) ) {
        mexErrMsgTxt("D!=(ds*nsq): Passed data is not compatible with the PQ structure. Make sure m is a multiple of data dimensionality.");
    }

    // output matrix , initialized to 0
    plhs[0] = mxCreateNumericMatrix(m, N, mxUINT8_CLASS, mxREAL);
    uint8_t *out = (uint8_t *) mxGetData(plhs[0]);

    process(ds, ks, m, N, X, cent3s, out );

}


