/*
% INPUT
% codes             (sum(nsq_z) N)  uint8   codes where nsq_z = nsq_c(z2c). (Base 0)
% l2normsq_c         (1 C)           cell    Each one; (ks nsq_c(c)),double. Obtained via {pq.l2normsq}.
% z2c               (1 Z)           uint64  Defines sharing of PQ codes across subvector index by z. (Base 1)
%
% OUTPUT
% out               (Z N)           double  Squared-norm of each subvector. 
*/

/**
 * Author: R. Gokberk Cinbis.
 */


#include "mex.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <cstring>
#include <stdint.h> // to get uint8_t
#include <limits>
#include <cstdarg>
#include "cn_matlabarrays.hpp"

#include "omp.h"
#include "cn_threads.hpp"

using namespace std;

#define ks 256

enum {a_codes=0,a_l2normsq_c,a_z2c};
void mexFunction(int nlhs, mxArray *plhs[],int nrhs, const mxArray *prhs[]) 
{
    if ( (nlhs!=1) || (nrhs!=3) )
        mexErrMsgTxt("invalid number of input/output arguments");

    // sizes
    mwSize sum_nsq_z = cn::mex::getdim(prhs[a_codes],1);
    mwSize N = cn::mex::getdim(prhs[a_codes],2);
    mwSize C = mxGetNumberOfElements(prhs[a_l2normsq_c]);
    mwSize Z = mxGetNumberOfElements(prhs[a_z2c]);

    // pointers
    const uint8_t *codes = cn::mex::getDenseData<uint8_t>(prhs[a_codes]);
    vector< const double *> l2normsq_c;
    vector< mwSize > nsq_c;
    l2normsq_c.reserve(C);
    nsq_c.reserve(C);
    if (!mxIsCell(prhs[a_l2normsq_c]))
            mexErrMsgTxt("l2normsq_c is not a cell array");

    for (mwIndex c = 0; c < C; ++c) {
        const mxArray *a = mxGetCell(prhs[a_l2normsq_c],c);
        if (a==NULL)
            mexErrMsgTxt("l2normsq_c contains NULL pointers");
        if (mxGetM(a)!=ks)
            mexErrMsgTxt("ks!=256");
        nsq_c.push_back( mxGetN(a) );
        l2normsq_c.push_back( cn::mex::getDenseData<double>(a) );
    }
    const uint64_t *z2c = cn::mex::getDenseData<uint64_t>(prhs[a_z2c]);

    // sanity checks
    {
        mwSize sum_nsq_z2= 0;
        for (mwIndex i = 0; i < Z; ++i) {
            const uint64_t &c = z2c[i] - 1;
            if ( c >= C )
                mexErrMsgTxt("there are invalid entries in z2c");
            sum_nsq_z2 += nsq_c[c];
        }
        if (sum_nsq_z2!=sum_nsq_z) {
            printf("size(codes,2)=%d, sum(nsq_c(z2c))=%d\n",sum_nsq_z,sum_nsq_z2);
            mexErrMsgTxt("inconsistent sum(nsq_z)");
        }
    }

    // shortcuts
    vector< mwSize > nsq_z; 
    nsq_z.reserve(Z);
    for (mwIndex z = 0; z < Z; ++z) {
        nsq_z.push_back( nsq_c[ z2c[z] - 1 ] );
    }

    vector< const double * > sqnorms_z;
    sqnorms_z.reserve(Z);
    for (mwIndex z = 0; z < Z; ++z) {
        sqnorms_z.push_back( l2normsq_c[ z2c[z]-1 ] ); 
    }

    // allocate output
    mwSize outdims[2] = {Z,N};
    plhs[0] = mxCreateNumericArray(2,outdims,mxDOUBLE_CLASS,mxREAL);
    double * out = (double *) mxGetData(plhs[0]);

#if 0
    // compute
    if (N > 0)
        omp_set_num_threads(MIN(cn::setmaxthreads(),N));
    else
        omp_set_num_threads(1);

#pragma omp parallel for
    for (mwIndex i = 0; i < N; ++i) {

#else
    // much faster !!!!
    omp_set_num_threads(1);
    for (mwIndex i = 0; i < N; ++i) {
#endif


        double *xout = out + i * Z;
        const uint8_t *xcode = codes + sum_nsq_z * i;

        for (mwIndex z = 0; z < Z; ++z) {

            const double *xnorms = sqnorms_z[z]; // points to (ks nsq_z) matrix
            double curout = 0;
            const mwSize cur_nsq = nsq_z[z];
            for (mwIndex j = 0; j < cur_nsq; ++j) {
                curout += xnorms[ *(xcode++) ];
                xnorms += ks;
            }

            *(xout++) = curout;
        } // over z

    } // over i
    

}
    

