#pragma once

/**
 * Author: R. Gokberk Cinbis.
 */


#define FvGmm_PoolObjDesc_PtWeightEps 1e-20f

// #define DEBUG_SPATFV

// NOTES
// *** We assume that the number of objects is significantly high so that we can multithread over the windows. 
// * Floating data is in double precision for simplicity.
// * There are at most 256 scales [which is I think is definitely fine!].
// * We do not support d_rows.' * posterior, since it seems to be slow in cn_mtimesinnermask()
// * By taking points as cell arrays over scales, we help faster processing, reducing bb-point comparisons.
// * Keep mwIndex as it is. For non-matlab usage, simply typedef it to size_t.
//
// R.G.Cinbis, Sep 2012

// todo: separate segmentation stuff to another class.

#include "mex.h"
#include <omp.h>
#include <cmath>
#include <math.h>
#include <vector>
#include <algorithm>
#include <cstring>
#include "cn_threads.hpp"
#include "cn_math.hpp"
#include "cn_memory.hpp"
#include "cn_trees.hpp"
#include <vector>
#include <iostream>
#include <string>
#include <cn_gaussmixture.hpp>
#include <sstream>

using std::vector;
using std::string;
using std::runtime_error;

namespace cn { namespace fvgmm {

    /** 
     * Shared utility functions.
     */
    namespace utils {

        /// Do to += from.
        template <class T> static void addToVector(T *from, T *to, mwSize len);

        /**
         * Utility function to print a matrix.
         */
        template <class TT> 
            static void printMatrix( const uint64_t M, const uint64_t N, const TT * vals, const char *name );

        template <class TT>
            string num2str(const TT i);

        /// Return sqrt(max(val,eps)) assuming val is non-negative.
        double mysqrt(const double &val) ;

        // Set fisher vector normalizer: (1+2*D K)
        void setfvnormalizer(const mwSize K,const mwSize D,const double *mix,const double *vars,double *fvnormalizer);

#include "utils_templates.cpp"
    }

    /**
     * Embedding these into PoolObjDesc makes poolobjdesc_templates.cpp very complicated.
     * We better share them.
     */
    typedef uint32_t id_t; 
    typedef double npts_t; 

    // parameters necessary for the main computations. keep this matlab-independent. (mwSize, mxAlloc, etc are fine
    // since easy to re-define)
    //
    // Tc should usually be int16_t since most images have size larger than 256.
    // the only case where Tc uint8_t may really be useful is that using boxes only (no segmentation)
    // and using uint8-normlized box and point coordinates.
    template <class Tc> class PoolObjDesc {

        public:

            enum MaskType { MaskNone, MaskBool, MaskFloat, MaskBackground, MaskImage };

            /**
             * Constructor.
             */
            PoolObjDesc();
            /**
             * Pool descriptors.
             * Must be preallocated: out: (D K R), npts: (1 R)
             * out should have been pre-initialized to zeros.
             */
            void poolboxdesc(double *out,npts_t *npts) ;

            /**
             * Pool Fisher Vectors.
             * Must be preallocated: out: (1+2*D K R), npts: (1 R)
             * out should have been pre-initialized to zeros.
             *
             * If avgfv=true, fv descriptor is normalized by the number of points in the box.
             */
            void poolboxfv(double *out,npts_t *npts,bool avgfv) ;

            /**
             * Pool fishervectors according to a segmentation tree.
             * out should be pre-allocated and pre-initialized to zeros.
             * No descriptor normalization is done.
             */
            void pooltreefv(double *out,npts_t *npts,shared_ptr<const id_t> idmap,const Tc imH,const Tc imW,const SingleTree <id_t> &tree) ;


            /**
             * Pool Spatial Fisher Vectors.
             * Must be preallocated: out: (1+4 K*Kapp R), npts: (1 R)
             * out should have been pre-initialized to zeros.
             */
            void poolboxspatfv(double *out,npts_t *npts,bool avgfv) ; // poolboxspatfv


            /**
             * Pool Spatial Fisher Vectors over a segmentation tree.
             * Must be preallocated: out: (1+4 K*Kapp R), npts: (1 R)
             * out should have been pre-initialized to zeros.
             * No descriptor normalization is done.
             *
             * TODO This is somewhat slow. To speed up there are things that we can do:
             *      - For example, we can move information from lower levels to higher levels in order to find out
             *      points associated with each segment.
             *      - Similarly, we can also find out the bounding box of the upper level pretty much the same
             *      way.
             *      - Therefore, there is a lot to be done if it becomes necessary to speed up!
             */
            void pooltreespatfv(double *out,npts_t *npts,shared_ptr<const id_t> idmap,const Tc imH,const Tc imW,const SingleTree <id_t> &tree) ;

            /// Set dictionary parameters. This is not needed for just pooling without FV computation.
            void setdictionary(const mwSize D,const mwSize K,const double *mix,const double *means,const double *vars) ;

            /// Forget any existing dictionary info.
            void unsetDictionary() ;

            mwSize getnumregions() ;

            /**
             * Set boxes for box-based pooling.
             * bbs: (4 R), base 0 coordinates.
             */
            void setboxes(shared_ptr<const Tc> bbs,const mwSize R) ;

            /**
             * Remove any masks definitions done via setmasks().
             */
            void unsetmasks() ;

            /**
             * Set per-box binary segmentation mask.
             * If this is defined, any method utilizing getptmask4box will be affected.
             */
            void setmasks(const shared_ptr< vector< shared_ptr<const bool> > > &masks ) ;


            /**
             * Set per-box segmentation weights. Each point will be weighted based on its weight.
             * If this is defined, any method utilizing getptmask4box will be affected.
             */
            void setmasks(const shared_ptr< vector< shared_ptr<const float> > > &masks ) ;

            /**
             * setmasks(MaskBackground)
             *      Pool descriptors that are strictly outside the bounding boxes.
             */
            void setmasks(const MaskType &masktype) ;

            /**
             * setmasks(maskim,height,width,nbank,maskind)
             *      Use maskim's maskind[i]-th band for masking the i-th box.
             *      maskind shoud be base 0, in size (1 R)
             *      maskim should be col-major.
             */
            void setmasks(const shared_ptr<const float> maskim,size_t maskim_H, size_t maskim_W, size_t maskim_B,const shared_ptr<const int16_t> maskind);

            /// make sure this is called after setboxes() or setsegmentation().
            void setscales(const int16_t *minscs,const int16_t *maxscs,const int16_t scstep) ;

            /// remove scale information
            void unsetscales() ;

            /// add a new scale of sparse posterior of shape (N Kapp).
            /// note that in spatial fv, this->K = Kspatial, which is different from Kapp.
            void pushPosteriors( mwSize N, mwSize Kapp, const double *Q, const mwIndex *Ir, const mwIndex *Jc ) ;

            // add a new scale of locations. locations should be base 0.
            void pushLoc( mwSize N, shared_ptr< const Tc > fxy ) ;

            // add a new scale of appearances
            void pushDesc( mwSize Dapp, mwSize N, const double * d_cols ) ;

            // call this immediately after adding posteriors,locations,[descriptors]
            void finalizePatchInfo();

            mwSize getnscale() ;

            mwSize getDapp() ;

            mwSize getKapp(); 

            mwSize getKdict();

            // it is just practical to leave some fields public.
            // these should be filled by the user.

        private:

            // -- patch info --
            bool patchfinal;
            vector<const double *> allQ;    // sparse posteriors.
            vector<const mwIndex *> allQ_Ir;// sparse posteriors.
            vector<const mwIndex *> allQ_Jc;// sparse posteriors. 
            vector< shared_ptr<const Tc> > allLocs;     // feature locations. just like bbs, this should be base 0.
            vector<const double *> allV; 
            vector<mwSize> allN;            // number of points.
            mwSize _nscale;                // requires patchfinal, therefore, never access directly.
            mwSize Kapp;         // based on posterior probabilities.
            mwSize Dapp;         // based on descriptors. if spatial fv is being computed, then Dapp can be invalid.

            // -- dictionary info --
            mwSize D;            // dictinary's D.
            mwSize K;            // dictionary's K.
            const double *mix;   // (1 K) mixing weights
            const double *means; // (D K)
            const double *vars;  // (D K)
            shared_ptr<double>fvnormalizer; // (1+2*D K)
            shared_ptr<double>meansSq;

            // -- box info --
            mwSize R;                           // either number of regions (boxes or segments)
            shared_ptr<const Tc> bbs;           // boxes, base 0. (4 R).
            shared_ptr <bool> segmentexists;    // segmentexists[r] = any( idmap==(r+1) )

            // -- mask within the each segment --
            MaskType masktype;
            shared_ptr< vector< shared_ptr<const bool> > > masks_bool; // per-box segmentation masks
            shared_ptr< vector< shared_ptr<const float> > > masks_float; // per-box segmentation weights

            // masktype: MaskImage
            shared_ptr<const float> maskim;       // full-image multi-band mask image (H W B)
            shared_ptr<const int16_t> maskind; // band indices
            size_t maskim_H;                      // height
            size_t maskim_W;                      // width
            size_t maskim_B;                      // band

            // -- segmentation tree --
            // segment2pts[ segment2pts_index[i*nscale + sci] + j ] gives j-th point at sci-th scale
            //      and within the segment id=(i+1). That is, result is a point index within the corresponding scale.
            shared_ptr <mwSize> segment2pts;            // concatenatation of index arrays.
            shared_ptr <mwSize> segment2pts_index;      // (nscale R)

            // -- pooling scales --
            const int16_t *minscs; // base 0!
            const int16_t *maxscs; // base 0!
            int16_t scstep;

            /**
             * Get min..max inclusive scale for a particular region.
             */
            void getMinMaxScaleForRegion( mwSize r, int16_t *minscale, int16_t *maxscale ) ;

            /**
             * A wrapper to omp_get_thread_num()
             */
            mwSize getthreadindex( const mwSize & maxthread ) const ;

            /**
             * Add new N (if needed) to allN vector. 
             */
            void pushN( mwSize N, mwSize sci ) ;

            /// Create (2D K) matrices containing [d_cols; d_cols.^2] 
            void catWithSquares(vector< shared_ptr<double> > *allV2) ;

            /**
             * Convert 0-th,1-th and 2-nd order cumulated posteriors, descriptors, descriptor^2 (i.e.
             * moments*npts) into normalized fisher vectors with analytical approximation to Fisher Information
             * Matrix.  This function works on the FV of a single object.
             *
             * If there are no points inside, the obj_out will remain unchanged, assuming it is all-zeros.
             */
            void stats2fv(double *obj_out,double obj_npts,bool avgfv) ; // stats2fv()

            /// Set number of threads. Return maximum number of threads.
            mwSize setnumparallel( mwSize njob ) ; // setnumparallel()


            /// Single pass to find #points within each segment into (R nscale) matrix.
            shared_ptr<mwSize> countPtsPerRegionPerScale(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) ;

            /** 
             * Compare box cooodinates and points coordinates within the current scale
             *
             * sci          Feature scale index
             * r            Box index, base 0.
             * tmpm         Pre-allocated mask of length allN[sci]
             *
             * Returns sum(tmpm).
             */
            npts_t getptmask4box(const mwSize sci,const mwSize r,bool *tmpm) ;


            /**
             * Just like the bool* counter-part, but, for masktype==MaskFloat 
             * && masktype==MaskImage
             */
            double getptmask4box(const mwSize sci,const mwSize r,float *tmpm);

            /** 
             * Compute a binary mask for point lying inside the current segment.
             * Uses segment2pts and segment2pts_index.
             * 
             * sci          Feature scale index
             * r            Region index = (Region id - 1).
             * tmpm         Pre-allocated mask of length allN[sci]
             *
             * Returns sum(tmpm).
             */
            npts_t getptmask4segment(const mwSize nscale,const mwSize sci,const mwSize r,bool *tmpm) ;

            /**
             * Return total number of points within a particular segment. 
             * This is a very fast operation, utilizing the index structure.
             */
            /* -- not used, not tested --
             * npts_t getsegmenttotalnpts(const mwSize r) {
             mwIndex ii = nscale * r; // segment2pts_index: (nscale R)
             return segment2pts_index.get()[ii + nscale] - segment2pts_index.get()[ii];
             }*/


            /// SEE accumFVsToParents()
            void accumFVsToParentsHelper( double *out, npts_t *npts, const SingleTree <id_t> &tree, id_t root, mwSize fvlen ) ;

            /// max of a std::vector
            mwSize max( const vector<mwSize> &vec ) ;

            /// find region bounding boxes in base 0. requires R to be correct. 
            void idmap2bbs(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) ;

            /**
             * V:       (4 N) Normalized locations of points.
             * spatgmm
             * tmpm     (1 N) Denotes the points inside the box.
             * Qspat    (Kspat N) Output posterior probabilities. Qspat(:,~tmpm) contains garbage values.
             */
            template <class mask_t>
                static void getSpatialPosteriors( const uint64_t Npt, const double * V, const MixOfGauss<double> & spatgmm, const mask_t *tmpm, double * Qspat ) ;

            /**
             * Pool elements towards upper nodes in the tree.
             * This should be called after setsegmentation().
             *
             * Parents denotes the parent of each node where node indices have base 1.
             * Root node has parent=0. Root is required to be unique.
             *
             * In the "out" and "npts" arrays, leaf node descriptors are already computed with desctrans=none
             * Other nodes descriptors should be set to zeros before this function is called.
             */
            void accumFVsToParents( double *out, npts_t *npts, const SingleTree <id_t> &tree ) ;

            /** 
             * Pool fishervectors according to a segment ids image.
             * Algorithm:
             * 1) First pass to find #points within each segment.
             * 2) Second pass to find points within each segment.
             * 3) Pool them.
             *
             * (1) and (2) is already done in setsegmentation()
             *
             * out should have been pre-initialized to zeros.
             */
            void pooltreefv_leaves(double *out,npts_t *npts,bool avgfv,bool skipifnonexist) ; 


            /**
             * Pool Spatial Fisher Vectors over a segmentation image.
             * Must be preallocated: out: (1+4 K*Kapp R), npts: (1 R)
             * out should have been pre-initialized to zeros.
             */
            void pooltreespatfv_leaves(double *out,npts_t *npts,bool avgfv,bool skipifnonexist) ; 

            /**
             * Set segmentation map. 
             * R = max( idmap(:) ). idmap==0 regions denote "background".
             * Resets this->bbs to segment bounding boxes.
             */
            void setsegmentation(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) ;

            /**
             * Given the number of points within each region, craete "segment2pts_index" array.
             * Note that segment2pts_index does not contain the actual segment->pts index.
             * Instead, it just stores pointers to acces segment2pts array for each region.
             *
             * Returns total number of points in all regions, ie. sum(ptcounts)
             */
            mwSize make_segment2pts_index( const shared_ptr<mwSize> ptcounts ) ;


            /**
             * Build up "segmentexists" vector.
             * Note that segmentexists[r] refers to idmap==(r+1)
             * ie. r is base 0 in segmentexists and base 1 in idmap.
             * Requires "R" in accordance with idmap.
             */
            void setSegmentExists(shared_ptr<const uint32_t> idmap,const Tc imH,const Tc imW) ;


    };

    /**
     * Utility to compute single-patch FVs.
     */
    class PatchDesc {

        public:

            /**
             * Constructor.
             */
            PatchDesc();

            /// Set/overwrite dictionary parameters. 
            void setdictionary(const mwSize D,const mwSize K,const double *mix,const double *means,const double *vars) ;

            /** 
             * Get FV of a single patch. Thread-safe.
             *
             * desc     (D 1) Patch Descriptor.
             * sp       Sparsity of the posteriors for the current point.
             * q        (sp 1) Sparse posterior values.
             * qind     (sp 1) Corresponding component indices of the sparse posteriors.
             * out      (2D+1 K) Preallocated and zero-initialized col-major output array.
             *
             * ** out should have been pre-initialized to zeros **
             */
            void getpatchfv_full(const double *desc,const mwSize sp,const double *q,const mwIndex *qind, double * out);

            /** 
             * Get FV of a single patch wrt a single component. Thread-safe.
             *
             * ARGUMENTS
             * desc     (D 1) Patch Descriptor.
             * q        (1 1) p(k|x)
             * k        The component for which we'll compute the FV.
             * out      (2D+1 1) Preallocated col-major output array. No need to zero-initialize.
             *
             * FISHER VECTORS WHEN p(k|x)=0:
             * - fv_mix = -pi / sqrt(pi * (1-pi))
             * - fv_mu and fv_var entries will be all zeros. ( "desc" does not matter at all. )
             */
            void getpatchfv_k(const double *desc,const double q, const mwSize k, double * out);


            /**
             * Equivalent to getpatchfv_k for p(k|x) = 1, slightly more efficient.
             * (ie. get FVs wrt the cluster to which the current "desc" is assigned.)
             * Thread-safe.
             *
             * ARGUMENTS
             * desc     (D 1) Patch Descriptor.
             * k        The component for which we'll compute the FV and for which p(k|x)=1.
             * out      (2D+1 1) Preallocated col-major output array. No need to zero-initialize.
             */
            void getpatchfv_hardassign(const double *desc,const mwSize k, double * out);
                    
        private:

            // -- dictionary info --
            mwSize D;               // dictinary's D.
            mwSize K;               // dictionary's K.
            const double *mix;      // (1 K) mixing weights
            const double *means;    // (D K)
            const double *vars;     // (D K)

            shared_ptr<double>fvnormalizer_sh; // (1+2*D K)
            double *fvnormalizer_ptr; // just pointer
    
            // precomputed fv_mix for p(k|x)
            shared_ptr<double>fv_mix_zeropost_sh; // (1 K)
            double *fv_mix_zeropost_ptr; // just pointer

            mwSize fvlen_k;         // = 1+2*D. K * fvlen_k gives the total FV length. 

    };


#include "poolobjdesc_templates.cpp"

} }

#undef FvGmm_PoolObjDesc_PtWeightEps 

