#include <stdio.h>
#include <math.h>
#include <string.h>
#include <cstdio>
#include "cn_threads.hpp"
#include <unistd.h>
#include <sys/types.h>
#include "cn_macros.hpp"
#include <stdexcept>

/**
 * Author: R. Gokberk Cinbis.
 */

// whether to let this to control MKL settings. couldnt get this working yet
// #define MKL

#ifdef MKL
// not compatible: #include "mkl/mkl_serv_mkl_service.h"
// used nm -g mkl.so and public APIs to get these:
void mkl_serv_mkl_set_num_threads( int ); 
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN 255
#endif

#ifndef SHMDIR
#error SHMDIR is required for compilation.
#endif

#if CN_WINDOWS
    typedef unsigned int __pid_t;
#endif

typedef struct {
    __pid_t pid;
    int numprocs;
    bool init;
    char path[MAXPATHLEN+1]; 
} myinfo_t;

/// Helper function to initialize to store myinfo.
myinfo_t &initinfo() {

    static myinfo_t myinfo = { getpid(), omp_get_num_procs(), false, "\0" };

    if (!myinfo.init) {
        if ( strlen(SHMDIR) >= MAXPATHLEN ) {
            // dont use mex calls!
            fprintf(stderr,"Too long path. Increse MAXPATHLEN in cpuset.cpp\n");
        }
        sprintf(myinfo.path,"%s%s",SHMDIR,"cpuset.txt");
        myinfo.init = true;
    }

    return myinfo;
}

namespace cn {

    const char *getcpusetpath() {
        const myinfo_t &myinfo = initinfo();
        return myinfo.path;
    }

    int setmaxthreads(int argnthread,bool echo) {
        return _cn_threads::setmaxthreadsgeneric(argnthread,echo);
    }
    
    int setmaxthreads_limit(int argmaxthread,bool echo) {
        if (argmaxthread<=0)
            throw std::runtime_error("setmaxthreadsauto::argmaxthread is required to be positive");
        return _cn_threads::setmaxthreadsgeneric(-1,echo,argmaxthread);
    }

    namespace _cn_threads {
        int setmaxthreadsgeneric(int argnthread,bool echo,int argmaxthread) {

            const myinfo_t &myinfo = initinfo();

            // default
            // int res = myinfo.numprocs; // #thread=#core
            int res = 0; // dyanmic with up to all cores

            if (argnthread != -1) {
                res = argnthread;  // this overrides everything below.
            }
            else
            {

                // open cpuset file
                FILE *fid = fopen(myinfo.path,"r");

                if (echo) {

                    // printf("OLD omp_get_max_thread(): %d \n",omp_get_max_threads());
                    // printf("OLD omp_get_dynamic(): %d \n",omp_get_dynamic());

                    printf("Current pid=%d\n",myinfo.pid);
                    printf("Number of processors=%d\n",myinfo.numprocs);
                    if (fid == NULL)
                        printf("Doesnt exist: %s\n",myinfo.path);
                    else
                        printf("Exists: %s\n",myinfo.path);

                }

                while (fid!=NULL) { // while instead of if in order to use "break" easily.

                    // try to reading the settings

                    int y,n;
                    // unsigned int x;
                    __pid_t x;

                    // read default #cpu
                    n = fscanf(fid,"%d",&y);
                    if (n < 1) 
                        break; // eof
                    res = y; // override

                    if (echo) {
                        printf("Read default setting=%d\n",y);
                    }

                    while(true) {

                        // read pid-specific #cpu
                        unsigned int xuint;
                        n = fscanf(fid,"%u %d",&xuint,&y);
                        x = xuint;
                        if (n < 2) 
                            break;
                        if (x==myinfo.pid) {
                            res = y; // override.

                            if (echo) {
                                printf("Read current pid-specific setting=%d\n",y);
                            }

                            break; // dont look any more
                        }


                    }

                    break;

                }

                if (fid!=NULL) {
                    fclose(fid);
                }


            }

            int isdyn;
            if (res < 0) {
                isdyn = 1;
                res = -res;
            } else if ( res == 0 ) {
                isdyn = 1;
                res = myinfo.numprocs;
            } else {
                isdyn = 0;
            }

            // make sure we select a reasonable #threads
            res = MIN( myinfo.numprocs, res );
            if (argmaxthread>0)
                res = MIN( argmaxthread, res );

#ifdef MKL
            mkl_serv_mkl_set_num_threads(res);
#endif 
            omp_set_dynamic(isdyn);
            omp_set_num_threads(res);


            if (echo) {
#ifdef MKL
                printf("omp_set_num_threads & mkl_serv_mkl_set_num_threads to %d with non-dynamic.\n",res);
#else
                printf("omp_set_dynamic to %d \n",isdyn);
                printf("omp_set_num_threads to %d \n",res);
                // printf("NEW omp_get_max_threads(): %d \n",omp_get_max_threads());
                // printf("NEW omp_get_dynamic(): %d \n",omp_get_dynamic());
#endif
            }

            return res;
        }

        uint64_t posdivceil(uint64_t x, uint64_t y) {
            return (x % y) ? x / y + 1 : x / y;
        }
    }

    uint64_t jobstarti(uint64_t njob,uint64_t nworker,uint64_t thi) {
        return _cn_threads::posdivceil(njob,nworker) * thi;
    }

    uint64_t jobendi(uint64_t njob,uint64_t nworker,uint64_t thi) {
        return MIN(_cn_threads::posdivceil(njob,nworker) * (thi+1),njob);
    }


}

////////////////// misc

/*
int count_cpu() {
    cpu_set_t set;
    sched_getaffinity(0, sizeof(cpu_set_t), &set);
    int i, count=0;
    for(i=0;i<CPU_SETSIZE;i++)
        if(CPU_ISSET(i, &set)) count++;
    if (count>1) count--;

    return count;
}
*/


