#include <stdexcept>
#include "cn_memory.hpp"
using std::runtime_error;

/**
 * Author: R. Gokberk Cinbis.
 */


/**
 * Tools for dealing with "tree" structures.
 */

namespace cn {

    /**
     * Stores information about a tree (with a unique root).
     * Node ids have base 1, that is, 1 <= nodei <= nnode is always the case.
     *
     * parents array denotes the parent of each node where node indices have base 1.
     * Root node has parent=0. 
     */
    template <class T> class SingleTree {
        public:

            /// Constructor
            SingleTree(shared_ptr <const T> parents2,const T nnode2)
                : parents(parents2), nnode(nnode2), 
                  nchilds( new T[nnode](), DeleterArray<T> ), 
                  childs( new T[nnode-1](), DeleterArray<T> ),
                  childs_index( new T[nnode+1](), DeleterArray<T> ),
                  depths( new T[nnode](), DeleterArray<T> ),
                  maxdisttoleaf( new T[nnode](), DeleterArray<T> ),
                  nleaf( 0 )
            {
                // zero initialized arrays!

                checkAndSetRoot();
                setNumChildren();
                findChildren();

                treeheight = 0;
                setdepths(root,0);

                setmaxdisttoleaf(root);
                // printMatrix(1,nnode,maxdisttoleaf.get(),"maxdisttoleaf");

            }

            /// Get total number of nodes.
            const T getNumNode() const {
                return nnode;
            }

            /// Return root node's index.
            const T getRoot()const{
                return root;
            }
            
            /// Return the number of leaf nodes in the tree.
            void getNumLeaf() const {
                return nleaf;
            }

            /// Return the number of children of a particular node 1 <= nodei <= nnode.
            const T getNumChildren(const T &nodei) const {
                checknodei( nodei );
                return nchilds.get()[nodei-1];
            }

            /**
             * Return children of a particular node 1 <= nodei <= nnode.
             * Output should NOT be dealloc by the user.
             * For both input and output, 1 <= nodei <= nnode.
             */
            T *getChildren(const T &nodei) const {
                checknodei( nodei );
                return childs.get() + childs_index.get()[nodei-1];
            }

            /**
             * Return the parent of a particular node.
             * For both input and output, 1 <= nodei <= nnode.
             */
            T getParent(const T &nodei) const {
                checknodei( nodei );
                return parents.get()[nodei-1];
            }

            /**
             * Return the "depth" of a node, which is length of the path from the root to the node.
             * Root has depth = 0.
             * Depth of a leaf node is equal to or less than the height of a tree.
             * A tree with just the root node has depth = height = 0.
             *
             * Input: 1 <= nodei <= nnode
             * Output: 0 <= depth <= height
             */
            T getNodeDepth(const T &nodei) const {
                checknodei( nodei );
                return depths.get()[nodei-1];
            }

            /**
             * Return the "height" of a tree, which is the maximum depth in the tree.
             * SEE getNodeDepth()
             */
            T getTreeHeight() const {
                return treeheight;
            }

            /**
             * Return maximum distance from a leaf node to a particular node.
             *
             * For example, a root may connect to a large subtree and a leaf node.
             *     In any case, getNodeMaxDistToLeaf(root) is equal to tree height.
             *     Only for a complete/full binary tree, maxdisttoleaf is equal to treeheight-nodedepth.
             * 
             * Input: 1 <= nodei <= nnode
             * Output: 0 <= depth <= height
             */
            T getNodeMaxDistToLeaf( const T &nodei) const {
                checknodei( nodei );
                return maxdisttoleaf.get()[nodei-1];
            }



        private:

            shared_ptr <const T> parents; // root has parent=0, other indices are in range 1...nnode
            const T nnode;
            shared_ptr <T> nchilds;
            shared_ptr <T> childs;
            shared_ptr <T> childs_index; // childs[ childs_index[i] <= j < childs_index[i+1] ] are the children of the node with id (i+1).
            shared_ptr <T> depths;
            shared_ptr <T> maxdisttoleaf;
            T root; // root index, base 1, as everything else.
            T nleaf;
            T treeheight;

            /**
             * Find number of children per node.
             */
            void setNumChildren() {
                // find number of children of each node
                const T *parents = this->parents.get();
                T *nchilds = this->nchilds.get();
                for (mwIndex i = 0; i < nnode; ++i ) {
                    if ((i+1) != root)
                        ++(nchilds[parents[i]-1]);
                }
            }

            /**
             * Find the root node of the tree.
             * Throws exception
             * - if there are multiple root nodes.
             * - parents array contain invalid indices
             */
            void checkAndSetRoot() {
                bool found = false;
                const T *parents = this->parents.get();
                for (T i = 0; i < nnode; ++i) { 
                    if (parents[i] > nnode)
                        throw runtime_error("parents array contain out-of-bounds indices");
                    if (parents[i]==0) { 
                        if (!found) {
                            found = true;
                            root = i+1; // base 1 
                        } else {
                            throw runtime_error("SingleTree requires a unique root");
                        }
                    }
                }
            }

            /// Call this after setNumChildren
            void findChildren() {

                const T *parents = this->parents.get();
                const T *nchilds = this->nchilds.get();
                T *childs = this->childs.get();
                T *childs_index = this->childs_index.get();

                {
                    T n = 0;
                    childs_index[0] = 0;
                    for (mwIndex i = 0; i < nnode; ++i ) {
                        n += nchilds[i];
                        childs_index[i+1] = n;
                    }
                }

                shared_ptr<T> nchilds2_ptr( new T[nnode](), DeleterArray<T> ); // number of children added so far. init to zero.
                T *nchilds2 = nchilds2_ptr.get();
                for (mwIndex i = 0; i < nnode; ++i ) {
                    if ((i+1)!=root) {
                        const T &p = parents[i] - 1;
                        T &c = nchilds2[p];
                        childs[ childs_index[p] + c ] = i+1; // base 1
                        ++c;
                    }
                }
            }

            /**
             * Recursively set the depth of each node, starting from the root, which should have depth=0.
             * Also sets "nleaf".
             * cur: base 1, as always.
             */
            void setdepths(const T &cur,const T &depth) {
                depths.get()[cur-1] = depth;
                if (depth > treeheight)
                    treeheight = depth;
                T nc = getNumChildren(cur);
                if (nc > 0) { // non-leaf node
                    T *chld = getChildren(cur); // base 1
                    for (T i = 0; i < nc; ++i) {
                        setdepths( chld[i], depth+1 );
                    }
                } else {
                    ++nleaf;
                }
            }

            // Make sure that 1 <= nodei <= nnode
            void checknodei(const T &nodei) const {
                if ( (nodei > nnode) || (nodei < 1) )
                    throw runtime_error("0 <= nodei < nnode is not satisfied in getNumChildren()");
            }


            // Fill in "maxdisttoleaf"
            T setmaxdisttoleaf(const T &cur) {
                T nc = getNumChildren(cur);
                T *chld = getChildren(cur); // base 1
                T tmp = 0; // leaf has maxdisttoleaf=0
                for (T i = 0; i < nc; ++i) {
                    T tmp2 = setmaxdisttoleaf( chld[i] ) + 1;
                    if (tmp < tmp2) 
                        tmp = tmp2;
                }
                maxdisttoleaf.get()[cur-1] = tmp;
                return tmp;
            }

            /**
             * Utility function to print a matrix.
             */
            template <class TT> 
            static void printMatrix( const uint64_t M, const uint64_t N, const TT * vals, const char *name ) {

                printf("%s: ",name);
                if (M >= 2)
                    printf("\n");
                for ( uint64_t i = 0; i < M; ++i) {
                    for ( uint64_t j = 0; j < N; ++j) {
                        if ( std::numeric_limits<TT>::is_integer )
                            printf("%2d ",(int) vals[M * j + i]);
                        else
                            printf("%4.2f ",(double) vals[M * j + i]);
                    }
                    printf("\n");
                }
                printf("\n");

            }

    };



}
