// Implements function templates. Included directly by cn_pq.hpp

/**
 * Author: R. Gokberk Cinbis.
 */


template <class T> void squaredl2norm( 
        const vector<uint64_t>  &nsq_c,
        const vector<const T *> &sqnorms,
        const vector<uint64_t>  &z2c, 
        const uint64_t nx, 
        const uint8_t *X,
        const bool perz,
        T *out) {

    // check input
    const uint64_t Z = z2c.size();
    const uint64_t C = nsq_c.size();
    if ( sqnorms.size()!=C ) 
        throw std::runtime_error("sqnorms vector has incompatible length");

    const uint64_t nsq_total = get_nsq_total(nsq_c,z2c);

    if ( perz ) {
#pragma omp parallel for
        for ( uint64_t xi = 0; xi < nx; ++xi ) {
            T *out_ = out + Z * xi;
            const uint8_t *x_ = X + nsq_total * xi;
            for ( uint64_t z = 0; z < Z; ++z ) {
                const uint64_t & c = z2c[z];
                uint64_t nsq_ = nsq_c[c];
                const T *sqnorms_ = sqnorms[c]; // (ks nsq_)
                for ( uint64_t sqi = 0; sqi < nsq_; ++sqi ) {
                    *out_    += sqnorms_[ *(x_++) ];
                    sqnorms_ += ks;
                }
                out_++;
            }
        }
    } else {
#pragma omp parallel for
        for ( uint64_t xi = 0; xi < nx; ++xi ) {
            T *out_ = out + xi;
            const uint8_t *x_ = X + nsq_total * xi;
            for ( uint64_t z = 0; z < Z; ++z ) {
                const uint64_t & c = z2c[z];
                uint64_t nsq_ = nsq_c[c];
                const T *sqnorms_ = sqnorms[c]; // (ks nsq_)
                for ( uint64_t sqi = 0; sqi < nsq_; ++sqi ) {
                    *out_    += sqnorms_[ *(x_++) ];
                    sqnorms_ += ks;
                }
            }
        }
    }

}

template <class T> void sqeucdist (
        const vector<uint64_t>  &nsq_c,
        const vector<const T *> &sqeucs,
        const vector<uint64_t>  &z2c, 
        const uint64_t nx, 
        const uint64_t ny, 
        const uint8_t *X,
        const uint8_t *Y,
        T *D) 
{
    const uint64_t npair = nx * ny;
    const uint64_t C     = nsq_c.size();
    const uint64_t Z     = z2c.size();
    const uint64_t ntril = ks * (ks+1) / 2;
    if ( sqeucs.size()!=C ) 
        throw std::runtime_error("sqeucs vector has incompatible length");

    const uint64_t nsq_total = get_nsq_total(nsq_c,z2c);

#pragma omp parallel for 
    for ( uint64_t outi = 0; outi < npair; ++outi) {

        const uint64_t yi = outi / nx;
        const uint64_t xi = outi - nx * yi;

        // get data
        const uint8_t *_x = X + nsq_total * xi;
        const uint8_t *_y = Y + nsq_total * yi;
        T & _d = D[outi];

        // get distance
        for ( uint64_t z = 0; z < Z; ++z ) {
            const uint64_t & c = z2c[z];
            uint64_t _nsq      = nsq_c[c];
            const T * _lk      = sqeucs[c];
            for ( uint64_t j = 0; j < _nsq; ++j ) {
                _d  += _lk[ cn::trilsub2ind(ks,*(_x++),*(_y++)) ];
                _lk += ntril;
            }
        }

    }
}

template <class T> void dotprod(
        const vector<uint64_t>  &nsq_c,
        const vector<const T *> &dotprods,
        const vector<uint64_t>  &z2c, 
        const uint64_t nx, 
        const uint64_t ny, 
        const uint8_t *X,
        const uint8_t *Y,
        T *D) {

    // does exactly the same operations needed with a different lookup table.
    sqeucdist (nsq_c,dotprods,z2c,nx,ny,X,Y,D);
}


template <class T,class TOp> void _::shared_dist_l2normperz(
        const vector<uint64_t>   &nsq_c,
        const vector <const T *> &dotprods,
        const vector <const T *> &sqnorms,
        const vector<uint64_t>   &z2c, 
        const uint64_t nx, 
        const uint64_t ny, 
        const uint8_t *X,
        const uint8_t *Y,
        T *D) {

    const uint64_t npair = nx * ny;
    const uint64_t C     = nsq_c.size();
    const uint64_t Z     = z2c.size();
    const uint64_t ntril = ks * (ks+1) / 2;
    if ( dotprods.size()!=C ) 
        throw std::runtime_error("sqnorms vector has incompatible length");

    const uint64_t nsq_total = get_nsq_total(nsq_c,z2c);

    // first compute the norm of each vector as (Z N)
    shared_ptr<T> Xnorms_ptr( new T[Z*nx](), DeleterArray<T> ); // zero initialized
    shared_ptr<T> Ynorms_ptr( new T[Z*ny](), DeleterArray<T> ); // zero initialized
    T *Xnorms = Xnorms_ptr.get();
    T *Ynorms = Ynorms_ptr.get();
    squaredl2norm(nsq_c,sqnorms,z2c,nx,X,true,Xnorms);
    squaredl2norm(nsq_c,sqnorms,z2c,ny,Y,true,Ynorms);
    _::sqrtmax(Z*nx,Xnorms_ptr,(T) 1e-30);
    _::sqrtmax(Z*ny,Ynorms_ptr,(T) 1e-30);

    // _::printmat("Xnorms",Z,nx,Xnorms);
    // _::printmat("Ynorms",Z,ny,Ynorms);

#pragma omp parallel for 
    for ( uint64_t outi = 0; outi < npair; ++outi) {

        const uint64_t yi = outi / nx;
        const uint64_t xi = outi - nx * yi;

        // get data
        const uint8_t *_x = X + nsq_total * xi;
        const uint8_t *_y = Y + nsq_total * yi;
        const T *_normx   = Xnorms + Z * xi; // (Z 1)
        const T *_normy   = Ynorms + Z * yi; // (Z 1)
        T & _d            = D[outi];

        // process
        for ( uint64_t z = 0; z < Z; ++z ) {
            const uint64_t &c = z2c[z];
            uint64_t _nsq     = nsq_c[c];
            const T * _lk     = dotprods[c];

            T _dot            = 0; 
            for ( uint64_t j = 0; j < _nsq; ++j ) {
                _dot += _lk[ cn::trilsub2ind(ks,*(_x++),*(_y++)) ];
                _lk += ntril;
            }

            // normalize by ||x_z|| * ||y_z||.  1e-30 is good in both float and double.
            // do not use macros-based max(), which is very risky!
            _d += _dot / cn::max( (T) 1e-30, *(_normx++) * *(_normy++) );
        }

        // dotprod:     _d 
        // sqeucdist:   2*(Z-_d)
        shared_dist_l2normperz_op<T,TOp>(_d,Z);
    }
}

namespace _ { // mandatory for template specialization. see shared_dist_l2normperz.
    template <> inline void shared_dist_l2normperz_op<float,_::shared_dist_l2normperz_op_dotprod>(float &,const uint64_t ) {} // no op
    template <> inline void shared_dist_l2normperz_op<double,_::shared_dist_l2normperz_op_dotprod>(double &,const uint64_t ) {} // no op
    template <> inline void shared_dist_l2normperz_op<float,_::shared_dist_l2normperz_op_sqeucdist>(float &_d,const uint64_t Z) { _d = 2 * ( Z - _d ); }
    template <> inline void shared_dist_l2normperz_op<double,_::shared_dist_l2normperz_op_sqeucdist>(double &_d,const uint64_t Z) { _d = 2 * ( Z - _d ); }
}

template <class T> void _::sqrtmax(const uint64_t n, const shared_ptr<T> &vals,const T eps) {
    T * x = vals.get();
    for ( uint64_t i = 0; i < n; ++i ) {
        if ( x[i] < eps ) 
            x[i] = std::sqrt( eps );
        else
            x[i] = std::sqrt( x[i] );
        // printf("_[i]=%.5f\n",x[i]);
    }
}

template <class T> void _::printmat(const char *name,const uint64_t m, const uint64_t n, const T *mat) {
    printf("----%s----\n",name);
    for ( uint64_t i = 0; i < m; ++i ) {
        for ( uint64_t j = 0; j < n; ++j ) {
            printf("%.2f  ",(T) mat[m*j+i]);
        }
        printf("\n");
    }
    printf("------------\n");
}


template <class T> void linclslookup(
        const vector<uint64_t>    &ds_c,
        const vector<uint64_t>    &nsq_c,
        const vector <const T *>  &centers,
        const vector<uint64_t>    &z2c, 
        const uint64_t            ncls,  
        const T *                 allw,
        T *                       alls) {

    // C,Z,nsq_total
    uint64_t C = ds_c.size();
    uint64_t Z = z2c.size();
    if ( (nsq_c.size()!=C) || (centers.size()!=C) )
        throw std::runtime_error("linclslookup: inconsistent C");
    uint64_t nsq_total = 0;
    for ( uint64_t z = 0; z < Z; ++z ) {
        nsq_total += nsq_c[z2c[z]];
    }

    // flatten over all subquantizers in order to multithread
    vector<const T *> center_sq; // (1 nsq_total) {(ds ks)}  where ds=ds_sq[sqi]
    vector<uint64_t>  ds_sq;     // (1 nsq_total) 
    vector<const T *> w_sq;      // (1 nsq_total) {(ncls ds)} where ds=ds_sq[sqi]
    center_sq.reserve(nsq_total);
    ds_sq.reserve(nsq_total);
    w_sq.reserve(nsq_total);
    { 
        const T *w = allw;
        for ( uint64_t z = 0; z < Z; ++z ) {
            const uint64_t c = z2c[z];
            const uint64_t ds = ds_c[c];
            const uint64_t nsq = nsq_c[c];
            const T *cent = centers[c]; // (ds ks _nsq)
            for ( uint64_t s = 0; s < nsq; ++s ) {
                center_sq.push_back(cent);
                ds_sq.push_back(ds);
                w_sq.push_back(w);
                cent  += ds * ks;
                w     += ncls * ds; 
            }
        }
    }

    // note: parfor is not necessary, this is already very fast.
    // however, if you ever hit to a problem where this is not fast enough, just put openmp pragma over this.
    uint64_t nclsks = ncls * ks;
    for ( uint64_t sqi = 0; sqi < nsq_total; ++sqi ) {
        const T *w = w_sq[sqi];         // (ncls ds)
        const T *cent = center_sq[sqi]; // (ds ks)
        const uint64_t ds = ds_sq[sqi];
        T *s = alls + (nclsks * sqi); // (ncls ks)

        // compute s = w * cent = (ncls ds)*(ds ks) = (ncls ks)
        cn::blas::mtimes('N',w,'N',cent,s,ncls,ds,ks);
    }

}

template <class T> void XTTmtimesW(
            const vector<uint64_t>    &ds_c,
            const vector<uint64_t>    &nsq_c,
            const vector<const T *>   &centers,
            const vector<uint64_t>    &z2c, 
            const uint64_t            N,  
            const uint64_t            K,  
            const uint8_t *           XT,
            const T *                 wscr,
            const T *                 W,
            T *                       out)
{

    // C,Z,nsq_total
    uint64_t C = ds_c.size();
    uint64_t Z = z2c.size();
    if ( (nsq_c.size()!=C) || (centers.size()!=C) )
        throw std::runtime_error("XTTmtimesW: inconsistent C");
    uint64_t nsq_total = 0;
    for ( uint64_t z = 0; z < Z; ++z ) {
        const uint64_t c = z2c[z];
        nsq_total += nsq_c[c];
        // d_total += nsq_c[c] * ds_c[c];
    }

    // We can safely assume that (ks d) is feasible to store in memory.
    // If this assumption is wrong, then the output matrix (D d) has to be infesible since ks << D!

    // expand centers to simplify later ops.
    // subquantizer sqi (base 0) corresponds to the dimension block [cumds_sq[sqi],cumds_sq[sqi+1]).
    vector<const T *> center_sq;  // (1 nsq_total) {(ds ks)}  where ds=ds_sq[sqi]
    vector<uint64_t> cumds_sq; // (1 nsq_total+1) 
    vector<uint64_t> sq2z; // (1 nsq_total) 
    center_sq.reserve(nsq_total);
    cumds_sq.reserve(nsq_total+1);
    sq2z.reserve(nsq_total);
    cumds_sq.push_back(0);
    { 
        for ( uint64_t z = 0; z < Z; ++z ) {
            const uint64_t c = z2c[z];
            const uint64_t ds = ds_c[c];
            const uint64_t nsq = nsq_c[c];
            const T *cent = centers[c]; // (ds ks nsq)
            for ( uint64_t s = 0; s < nsq; ++s ) {
                center_sq.push_back(cent);
                sq2z.push_back(z);
                cumds_sq.push_back(cumds_sq.back()+ds);
                cent += ds * ks;
            }
        }
    }
    uint64_t d_total = cumds_sq.back(); // = cumds_sq[nsq_total].

    /*
     * I've considered two alternatives for multithreading here:
     *
     * (A) Multithread over (subquantizer sqi,k) pairs. In each worker, we create a weight histogram (ks 1)
     * over the centers by processing XT(:,sqi) and W(:,k). Then a (ds_z ks)(ks 1)=(ds_z 1) gives the sub-block in
     * out(:,k). Here, total temp array size is ks*nthread, which is almost always fine.
     * Assuming that N is large, it is important to keep XT local as much as possible within threads.
     *
     * (B) Multithread over (subquantizer sqi) only. In the case, we'll pool a weight histogram (ks K)
     * by processing XT(:,sqi) and W. The main advantage of this approach is potentially faster matrix
     * multiplication and increased locality in processing XT. The main disadvantages are that we may not 
     * multiprocess well if K is really small (unlikely) and total temporary array size ks*K*nthread may become
     * infeasible.
     * 
     * Without further consideration, I choose (A) due to its memory-friendliness, though, it is hard 
     * to know whether it is really a better choice. 
     */

    uint64_t jobsize = K * nsq_total; 
    cn::setmaxthreads_limit(jobsize);

    bool iserr = false;
#pragma omp parallel
    {
        T * hist = NULL; // temporary NULL array of size (ks 1)
        try {
            hist = new T [ks];
            uint64_t job_start = cn::jobstarti(jobsize);
            uint64_t job_end = cn::jobendi(jobsize);
            for (uint64_t jobi = job_start; jobi < job_end; ++jobi) {
                // Compute out(cumds_sq[sqi]:cumds_sq[sqi+1]-1,j)
                // Think of a matrix (K nsq_total)
                uint64_t sqi = jobi / K;       // [0,nsq_total), slowly increments
                uint64_t _k = jobi - sqi * K;  // [0,K), rapidly increments

                // get data
                const T *_cent        = center_sq[sqi]; // (ds ks)
                const uint64_t _d_beg = cumds_sq[sqi];
                const uint64_t _ds    = cumds_sq[sqi+1] - _d_beg;
                const uint8_t *_x     = XT + N * sqi; // (N 1)
                const T *_w           = W + N * _k; // (N 1)

                // histogram
                memset(hist,0,ks * sizeof(T));
                if ( wscr == NULL ) {
                    for (uint64_t i = 0; i < N; ++i) {
                        hist[*(_x++)] += *(_w++);
                    }
                } else { 
                    const T * _wscr = wscr + N * sq2z[sqi]; // (N 1)
                    for (uint64_t i = 0; i < N; ++i) {
                        hist[*(_x++)] += *(_w++) * *(_wscr++);
                    }
                }

                // set output (ds ks)(ks 1)=(ds 1)
                T *_out = out + d_total * _k + _d_beg; // (ds 1)
                cn::blas::mtimes('N',_cent,'N',hist,_out,_ds,ks,1);
            }

        } catch ( ... ) {
            iserr = true;
        }

        if (hist!=NULL)
            delete [] hist;
    }

    if (iserr) {
        throw std::runtime_error("cn::XTTmtimesW::Error occurred during omp loop");
    }



}

template <class T> void wsumz(
        const vector<uint64_t>    &ds_c,
        const vector<uint64_t>    &nsq_c,
        const vector<const T *>   &centers,
        const vector<uint64_t>    &z2c, 
        const uint64_t            N,  
        const uint8_t *           X,
        const T *                 W,
        T *                       out) {

    // C,Z
    uint64_t C = ds_c.size();
    uint64_t Z = z2c.size();
    if ( (nsq_c.size()!=C) || (centers.size()!=C) )
        throw std::runtime_error("XTTmtimesW: inconsistent C");

    // nsq_total,d_z
    uint64_t d_z;
    uint64_t nsq_total = 0;
    for ( uint64_t z = 0; z < Z; ++z ) {
        const uint64_t c = z2c[z];
        nsq_total += nsq_c[c];
        if ( z == 0 )
            d_z = nsq_c[c] * ds_c[c];
        else if ( d_z != nsq_c[c] * ds_c[c] )
            throw std::runtime_error("cn::pq::wsumz::d_z has to be constant for all z");
    }

#pragma omp parallel for 
    for ( uint64_t i = 0; i < N; ++i ) {
        T *out_i = out + d_z * i; // (d_z 1) zeros initially.
        const uint8_t *x_ = X + nsq_total * i; // (nsq_total 1)
        const T *w_ = W + Z * i; // (Z 1)
        for ( uint64_t z = 0; z < Z; ++z ) {
            T * o_ = out_i; // (d_z 1) == (ds_c*nsq_c 1). each loop over z will increment out_i via o_.
            const T curw = *(w_++); // scalar W(z,i)
            const uint64_t c = z2c[z];
            const uint64_t ds = ds_c[c];
            const uint64_t nsq = nsq_c[c];
            const uint64_t dsks = ds * ks;
            const T *cent = centers[c]; // (ds ks nsq)
            for ( uint64_t sqi = 0; sqi < nsq; ++sqi ) {
                const T *c_  = cent + ds * *(x_++); // (ds 1)
                for ( uint64_t d = 0; d < ds; ++d) 
                    *(o_++) += curw * *(c_++);
                cent += dsks;
            }
        }
    }
}

