function [O,areaX,areaY] = cn_bboverlap(X,Y,option,use_matlab)
%O = bboverlap(X,Y,[option])
%
% Calculates the overlap between two bounding box sets in the way
% it is calculated in pascal voc challenge (option='union'):
% Overlap = Area_Intersection / Area_Union
% 
% INPUT
% X: m x 4, each row is (x1,y1,x2,y2), converted to double if not.
% Y: n x 4, each row is (x1,y1,x2,y2), converted to double if not.
% [option]: An extra option string:
%           'union'         Normalize intersection area wrt union of objects (default)
%           'small'         Normalize intersection area wrt small object.
%           'large'         Normalize intersection area wrt large object.
%           'none'          Just give the intersection.
%           'first'         Wrt. area of X.
%           'second'        Wrt. area of Y.
%           '<option>+cont' Any of the above definition + continuous coord. See (b).
% 
% OUTPUT
% O:        m x n, a (non-symmetric) matrix with entries as requested by 'option'.
% areaX
% areaY
%
% Notes:
% (a) Supports negative x,y values too. BBs with negative sizes are converted
%     into positive ones! So, there can be a problem if a bb is assumed to 
%     have size 0. No bb can have size 0, which does not make a lot of sense
%     as it will have overlap 0 with everything, even if they have the same
%     bb's -- which creates a problem in clustering with bb-overlap distance.
% (b) Bounding boxes are considered as images (pixel coordinates) by
%     default. Therefore, [0 0 0 0] is considered to have area 1, since it
%     constitutes a single pixel in the image. The height is defined as y1-y2+1.
%     When '+cont' option is used, "+1"s are disabled and coordinates are 
%     considered as real mathematical coordinates. Then, [0 0 0 0] would have
%     area 0.
%
% Oct 2009, fixed bug: cn_bboverlap([.5 .5 -.5 -.5],[.5 .5 -.5 -.5])
%
% R. Gokberk Cinbis, June 2009

% Implementation notes:
% - "use_matlab" option is just for debugging.
% - It is very important to implement sth both in mex and matlab because
%   cn_bboverlap_elwise() uses matlab version.

% $Id: cn_bboverlap.m,b 1.13 2013/02/05 16:37:25 cinbis Exp $

if nargin < 3
    option = [];
end

[X Y areaX areaY NORM_OPT IS_PIXELCOORD] = cn_bboverlap_prepare(X,Y,option);

if nargin > 3 && use_matlab
    % for sanity check
    O = cn_bboverlap_main_matlab(X,Y,areaX,areaY,NORM_OPT,IS_PIXELCOORD,false);
else
    % transposing first work little bit faster than working with row-wise matrices.
    % note that X,Y cannot be too large since O has size (m n).
    % todo: allow col-wise input.
    X = X.';
    Y = Y.';
    O = cn_bboverlap_main_mex(X,Y,areaX,areaY,NORM_OPT,IS_PIXELCOORD);
end


