
/**
 * Author: R. Gokberk Cinbis
 */



#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include "mex.h"
#include "cn_matlabarrays.hpp"
#include <omp.h>
#include "cn_threads.hpp"

#include "cn_bboverlap_functions.cpp"

#define COLWISE 1

void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    ////////////////////////////////////////////////
    // get inputs
    ////////////////////////////////////////////////

    if (nrhs != 6)
        mexErrMsgTxt("invalid #input\n");
    if (nlhs != 1)
        mexErrMsgTxt("invalid #output\n");

    const mxArray *X_arr = prhs[0];
    const mxArray *Y_arr = prhs[1];
    const mxArray *areaX_arr = prhs[2];
    const mxArray *areaY_arr = prhs[3];
    const NormType normopt = (NormType) (int) cn::mex::getscalar(prhs[4]);
    const bool pixelcoord = (bool) cn::mex::getscalar(prhs[5]);

#if COLWISE
    mwSize m = mxGetN(X_arr);
    mwSize n = mxGetN(Y_arr);
    if (mxGetM(X_arr)!=4) { mexErrMsgTxt("size(X,2)!=4"); }
    if (mxGetM(Y_arr)!=4) { mexErrMsgTxt("size(Y,2)!=4"); }
#else
    mwSize m = mxGetM(X_arr);
    mwSize n = mxGetM(Y_arr);
    if (mxGetN(X_arr)!=4) { mexErrMsgTxt("size(X,2)!=4"); }
    if (mxGetN(Y_arr)!=4) { mexErrMsgTxt("size(Y,2)!=4"); }
#endif

    const double *X = cn::mex::getDenseData<double>(X_arr);
    const double *Y = cn::mex::getDenseData<double>(Y_arr);
    const double *areaX = cn::mex::getDenseData<double>(areaX_arr);
    const double *areaY = cn::mex::getDenseData<double>(areaY_arr);

    if (normopt != NORM_NONE) {
        if ((mxGetNumberOfElements(areaX_arr)!=m) || (mxGetNumberOfElements(areaY_arr)!=n)) {
            printf("M=%d numel(areaX)=%d  N=%d numel(areaY)=%d\n",m,mxGetNumberOfElements(areaX_arr),n,mxGetNumberOfElements(areaY_arr));
            mexErrMsgTxt("areaX or areaY have incorrect size");
        }
    }

    ////////////////////////////////////////////////
    // process
    ////////////////////////////////////////////////
    
    // alloc output
    mxArray *O_arr = mxCreateDoubleMatrix(m,n,mxREAL); // INITIALIZED TO 0
    if (O_arr == NULL) {
        mexErrMsgTxt("Alloc error.");
    }
    double *O = (double *) mxGetData(O_arr);
    plhs[0] = O_arr;

    bool iserr = false;

    /*

    // I don't see any clear advantage in making this function multithreaded.
    // Here, I simply avoid unnecessarily high number of threads to avoid communication overhead.
    int nmaxthread = MAX( 1, m*n/5e5 ); 
    nmaxthread = MIN(nmaxthread,5);  // avoid too many threads
    // printf("nmaxthread=%d \n",nmaxthread);
    omp_set_num_threads( MIN( nmaxthread, cn::setmaxthreads() ) );
    
#pragma omp parallel for
    */

    // Outer loop should be over "y" to respect output matrix shape.

#if COLWISE
    if (normopt != NORM_NONE) {
        for (mwIndex yi = 0; yi < n; ++yi) {
            for (mwIndex xi = 0; xi < m; ++xi) {
                O[yi*m+xi] = getbboverlap(X+4*xi,Y+4*yi,areaX[xi],areaY[yi],pixelcoord,normopt,&iserr);
            }
        }
    } else { // dont compute area
        for (mwIndex yi = 0; yi < n; ++yi) {
            for (mwIndex xi = 0; xi < m; ++xi) { 
                O[yi*m+xi] = getbboverlap(X+4*xi,Y+4*yi,0,0,pixelcoord,normopt,&iserr);
            }
        }
    }
#else
    double cury[4];
    double curx[4];
    if (normopt != NORM_NONE) {
        for (mwIndex yi = 0; yi < n; ++yi) {
            for (mwIndex di = 0; di < 4; ++di) 
                cury[di] = Y[n*di+yi];
            for (mwIndex xi = 0; xi < m; ++xi) {
                for (mwIndex di = 0; di < 4; ++di)  // copying here has significant overhead. alternatively use "stride" inside "getbboverlap()"
                    curx[di] = X[m*di+xi];
                O[yi*m+xi] = getbboverlap(curx,cury,areaX[xi],areaY[yi],pixelcoord,normopt,&iserr);
            }
        }
    } else { // dont compute area 
        for (mwIndex yi = 0; yi < n; ++yi) {
            for (mwIndex di = 0; di < 4; ++di) 
                cury[di] = Y[n*di+yi];
            for (mwIndex xi = 0; xi < m; ++xi) { 
                for (mwIndex di = 0; di < 4; ++di) 
                    curx[di] = X[m*di+xi];
                O[yi*m+xi] = getbboverlap(curx,cury,0,0,pixelcoord,normopt,&iserr);
            }
        }
    }
#endif
    
    if (iserr) { 
        mexErrMsgTxt("Unknown normopt");
    }

}


