function data = cn_integralimget(intim,bbs,opts)
% data = cn_integralimget(intim,bbs,opts)
%
% For already created integral image by cn_integralimcreate,
% returns the features for the subwindows given. It is more 
% efficient to call this function at once with all windows.
%
% INPUT (SIMPLE)
% intim:   (m+1) x (n+1) x d1 x d2 x ...
% bbs:     nwnd x 4. Each row is x1,y1,x2,y2 in the original image.
%          The coordinates are used as the first and the last pixel locations.
% [opts]   'average'    Take the average of elements in each cell. 
%
% INPUT (BAND-SPECIFIC CELLS)
% intim    ... 
% bbs      nwnd x 4 x d1 x d2 x d3 x d4 x 1 x d5 x d6 ... 
%          Example: intim = (m+1) x (n+1) x K x S  and  bbs = (m+1) x (n+1) x K x S
%          Then, each intim(:,:,k,s) are retrieved using bbs(:,:,k,s)
%
%          bbs broadcasting:
%          For "1" dimensions in "bbs", singular dimension is automatically broadcasted so that for those dimensions, window
%          bounding box is not dimension-specific (integral image values are still dimension-specific).
%          Example: intim = (m+1) x (n+1) x K x S  and  bbs = (m+1) x (n+1) x 1 x S
%          Then, each intim(:,:,k,s) is retrieved using bbs(:,:,1,s)
%
%          intim broadcasting: 
%          For "1" dimensions in "intim", singular dimension is automatically broadcasted so that integral image is virtually
%          replicate alogn those dimensions (bounding boxes are still dimension-specific).
%          Example: intim = (m+1) x (n+1) x 1 x S  and  bbs = (m+1) x (n+1) x K x S
%          Then, each bbs(:,:,k,s) is retrieved from intim(:,:,1,s)
% [opts]   ...
%
% OUTPUT
% data:    nwnd x d1 x d2 x ... Each data is the sum of elements in the spatial area (without opts).
%
% SEE cn_integralimcreate, ndImIntegralImage 
%
% R.G.Cinbis Feb 2010

% todo: option to request a particular band per window?

cn_setvardefaults(true,'bbs',zeros(0,4),'opts',[]);
cn_assert(size(bbs,2)==4,'input bb matrix has invalid size');
nwnd = size(bbs,1);

nd = max([ndims(bbs),ndims(intim),3]); 
sz_intim = cn_sizend(intim,nd);
sz_bb = cn_sizend(bbs,nd);

intim_bandstep = sz_intim(1) * sz_intim(2);

if ndims(bbs)==2 
    % simple case

    nfeat = prod(sz_intim(3:end));

    data = zeros(nwnd,nfeat);
    j11 = ii_sub2ind(sz_intim,bbs(:,2),   bbs(:,1)  );
    j12 = ii_sub2ind(sz_intim,bbs(:,2),   bbs(:,3)+1);
    j21 = ii_sub2ind(sz_intim,bbs(:,4)+1, bbs(:,1)  );
    j22 = ii_sub2ind(sz_intim,bbs(:,4)+1, bbs(:,3)+1);
    for fi = 1:nfeat
        data(:,fi) = intim(j22) + intim(j11) - intim(j12) - intim(j21);
        if fi < nfeat
            j11 = j11 + intim_bandstep;
            j12 = j12 + intim_bandstep;
            j21 = j21 + intim_bandstep;
            j22 = j22 + intim_bandstep;
        end
    end

    if ischar(opts) && strcmp(opts,'average')
        % find areas
        w = bbs(:,3) - bbs(:,1) + 1;
        h = bbs(:,4) - bbs(:,2) + 1;
        ar = reshape(w .* h,[],1); % area
        data = bsxfun(@rdivide,data,ar);
    elseif ~isempty(opts)
        error('unknown opts');
    end

    data = reshape(data,[nwnd sz_intim(3:end)]);

else

    % this seem to work OK but needs more rigorous testing

    intimfeatsz = sz_intim(3:end);
    bbfeatsz = sz_bb(3:end); 

    % sanity check: any non-broadcasting dimension should match.
    x = bbfeatsz==intimfeatsz;
    x(bbfeatsz==1) = true;
    x(intimfeatsz==1) = true;
    cn_assert(all(x),'bbs array has incompatible size');

    % output size
    outfeatsz = max(intimfeatsz,bbfeatsz);
    nfeat = prod(outfeatsz);

    bbs = reshape(bbs,size(bbs,1),size(bbs,2),[]); % necessary
    data = zeros(nwnd,nfeat);
    for fi_out = 1:nfeat

        % convert integral image band index (fi) into bounding box band index (fi2)
        % here we handle bb specific and bb independent bands
        sub_out = ind2sub2(outfeatsz,fi_out);
        fi_bb = sub2ind2(bbfeatsz, min(sub_out,bbfeatsz) ); % broadcast if needed
        fi_intim = sub2ind2(intimfeatsz, min(sub_out,intimfeatsz) ); % broadcast if needed

        j11 = ii_sub2ind(sz_intim,bbs(:,2,fi_bb),   bbs(:,1,fi_bb)  ) + (fi_intim-1) * intim_bandstep;
        j12 = ii_sub2ind(sz_intim,bbs(:,2,fi_bb),   bbs(:,3,fi_bb)+1) + (fi_intim-1) * intim_bandstep;
        j21 = ii_sub2ind(sz_intim,bbs(:,4,fi_bb)+1, bbs(:,1,fi_bb)  ) + (fi_intim-1) * intim_bandstep;
        j22 = ii_sub2ind(sz_intim,bbs(:,4,fi_bb)+1, bbs(:,3,fi_bb)+1) + (fi_intim-1) * intim_bandstep;
        data(:,fi_out) = intim(j22) + intim(j11) - intim(j12) - intim(j21);

    end

    if ~isempty(opts)
        error('opts is not implemented for the generalized bb case');
    end

    data = reshape(data,[nwnd outfeatsz]);

end



function ind = ii_sub2ind(sz_intim,y,x)
% get coordinates in the first band

% ind = sub2ind(sz_intim,y,x);
ind = (x-1) .* sz_intim(1) + y;


