function d_cols = cn_localcolordesc(im,f_bb,pdesc)
% Extract Florent et al. - ECCV 2010 local color descriptors.
%
% INPUT
% im        (height width nbnd) Image, in single/double precision and in range [0 1].
% f_bb      (4 n)               Bounding boxes where color descriptor will be extracted.
%                               Each column is [x1 y1 x2 y2]' where 1 <= x <= imw, 1 <= y <= imh.
% pdesc
%   .grid                       Size of the grid. In ECCV, grid=[4 4].
%   .dtrans                     'none' | 'SqRoot' | 'SqRoot+L2' | 'L2'
%
% OUTPUT
% d_cols    (dlen n)
%
% Gokberk Cinbis 2012

% input check
assert(isfloat(im));
assert(min(im(:)) > -eps);

% truncate
imh = size(im,1);
imw = size(im,2);
nb = size(im,3);
f_bb(2,:) = max(f_bb(2,:),1);
f_bb(3,:) = min(f_bb(3,:),imw);
f_bb(4,:) = min(f_bb(4,:),imh);
n = size(f_bb,2);

ncell = prod(pdesc.grid);

% extract color descriptors
[davg,dstd] = cn_imlocalstats(struct2('im',im,'f_bb',f_bb,'grid',pdesc.grid));
% davg,dstd: each one is (npt 3 ncell), for rgb.
assert(size(davg,1)==n);

% NO! d_cols = [ reshape(davg,n,[]) reshape(dstd,n,[]) ]'; % (D npt)
d_cols = [ reshape(davg,n,nb*ncell) reshape(dstd,n,nb*ncell) ]'; % (D npt)
assert(size(d_cols,2)==n);

if ~isempty(d_cols)
    % truncate erronous non-positive values.
   % assert(min(d_cols(:))>-1e-5);
end
d_cols = max(d_cols,0);

switch(pdesc.dtrans)
    case 'none'
        % ok
    case 'SqRoot'
        d_cols = realsqrt(abs(d_cols)) .* sign(d_cols);
    case 'SqRoot+L2'
        d_cols = realsqrt(abs(d_cols)) .* sign(d_cols);
        d_cols = bsxfun(@rdivide,d_cols,eps+realsqrt(sum(d_cols.^2,1))); % normalize the concatenated vector, as in SIFT.
    case 'L2'
        d_cols = bsxfun(@rdivide,d_cols,eps+realsqrt(sum(d_cols.^2,1))); % normalize the concatenated vector, as in SIFT.
    otherwise
        error('unknown');
end

%- From Florent's ECCV2010 paper: "We also consider in some experiments simple
%96-D color features: a patch is subdivided into 4×4 sub-regions (as is the case
%of the SIFT descriptor) and we compute in each sub-region the mean and standard
% deviation for the three R, G and B channels. Both SIFT and color features are
%reduced to 64 dimensions using Principal Component Analysis (PCA)."

