function im = cn_rgb2gray(im)
% Convert RGB image into grayscale image. Equivalent to rgb2gray()
% but doesn't support colormap conversion.
% 
% INPUT
% im            (H W 3)
%
% OUTPUT
% im            (H W 1) The same class with the input.
%
% Warning: Based on rgb2gray()
%
% R.G.Cinbis, Oct 2011

persistent w

assert(ismember(class(im),{'double','single','uint8','uint16'}),'unsupported image data type');
assert(ndims(im)==3,'input should be HxWx3');
assert(size(im,3)==3,'input should be HxWx3');

% http://www.mathworks.com/help/toolbox/images/ref/rgb2gray.html:
% 
%   0.2989 * R + 0.5870 * G + 0.1140 * B
%
% but this is different around 1e-5 than what rgb2gray uses. normally, this is OK
% but for equivalency, we have to recalculate the inversion.

if isempty(w)
    x = inv([1.0 0.956 0.621; 1.0 -0.272 -0.647; 1.0 -1.106 1.703]);
    w = x(1,:)';
end

sz = size(im);
im = reshape(im,sz(1)*sz(2),sz(3));

switch(class(im))
    case {'double','single'}
        im = im * w;      
        im = min(max(im,0),1);
    case 'uint8'
        % eq: im = imlincomb(w(1),im(:,1),w(2),im(:,2),w(3),im(:,3),class(im));
        % elements exceeding the range are truncated end the rest is rounded (avoids overflow)
        % uint8 already does rounding.
        im = uint8( max( min( double(im) * w, 255 ), 0 ) );
    case 'uint16'
        im = uint16( max( min( double(im) * w, 65535 ), 0 ) );
    otherwise
        error('unsupported image data type');
end

im = reshape(im,sz(1:2));

