function [davg,dstd] = cn_imlocalstats(p)
% Extract statistics descriptors at provided local patch bounding boxes.
% Forms a spatial grid at each local patch bounding box and extracts mean and
% variance efficiently.
%
% This can be used to extract color descriptors of F. Perronnin ECCV 2010
% where:
%   grid = [4 4]
%   PCA is used to reduce dim from 96 to 64.
%
% INPUT
% p
%   .im         (H W D)     D-band image/feature maps.
%   .f_bb       (4 N)       Each column is a bounding box.
%   .grid       [nr nc]     Defines C=nr*nc cells.
%
% OUTPUT
% davg          (N D C)     Average descriptor.
% dstd          (N D C)     Standard deviation descriptor.
% 
% R.G.Cinbis Nov 2011

cellbb = cn_spatialgrid(p.f_bb,cellifnot(p.grid),struct2('normcoor',0)); % (4 C N)

% get (averaged) first and second moments
davg = helper(p.im,cellbb); % (N D C)
if nargout < 2
    return;
end

mom2 = helper(p.im.^2,cellbb); % (N D C)

% handle numerical negatives
x = detach(mom2)-davg.^2;
if ~isempty(x)
   % assert(min(x(:))>-1e-10);
end
x = max(x,0);

dstd = realsqrt(x);




function data = helper(im,bb)
% bb    (4 C N)
% 
% data  (N D C)

assert(size(bb,1)==4);
C = size(bb,2);
N = size(bb,3);
D = size(im,3);

bb2 = permute(bb,[3 1 2]); % (N 4 C)
intim = cn_integralimcreate(im);

for c = 1:C
    data(:,:,c) = cn_integralimget(intim,bb2(:,:,c),'average'); % (N D)
end
cn_assertsize([],data,N,D,C);



