function x = maskdim(x,d,m,shortcut)
% y = maskdim(x,d,m,shortcut)
%
% Safely apply a mask to a particular dimension of an array.
% Advantages over using matlab's version:
% - This is "safe", mask is forced to have proper length.
% - Has "shortcut check" option: If enabled and "all(m)==true", creating unnecessary copy of x is avoided.
%   This can mean huge speed-ups on large matrices.
% - Preserves ndims of x, avoiding "potential mistakes" like "x(:,m)" where x is 3-D.
% Disadvantage is that it has a small overhead as expected.
%
% USAGE
%
% y = maskdim(x,d,m)          
%   ->  y = x(:,...,m,...,:); % mask d-th dimension
%
% y = maskdim(x,d,m,true)
%   ->  y = x(:,...,m,...,:); % mask d-th dimension with "shortcut check".
%
% y = maskdim(x,':',m)          
%   ->  y = x(m); % mask over elements
% 
% R.Gokberk Cinbis, Jan 2012

% multithreading would probably bring no advantage.

assert(isscalar(d),'d has to be a scalar or '':''');

if nargin < 4
    shortcut = false;
end

% matlab doesn't check this: force mask to have the correct length
if d==':'
    % handle this specially
    
    checkmask(m,numel(x));

    if shortcut 
        % the behaviour of matlab here is pretty wild, just, don't bother supporting it.
        error('shortcut check is not support for d='':''');
    else
        x = x(m);
    end

    return;
end

checkmask(m,size(x,d)); 

if shortcut && all(m)
    % no change in size, y = x
    return;
end

sz = size(x);

if d == 1
    % handle faster

    x = x(m,:);
    sz(d) = size(x,d);
    x = reshape(x,sz);

elseif d == 2
    % handle faster

    x = x(:,m,:);
    sz(d) = size(x,d);
    x = reshape(x,sz);

elseif d == 3
    % handle faster

    x = x(:,:,m,:);
    sz(d) = size(x,d);
    x = reshape(x,sz);

elseif d == 4
    % handle faster

    x = x(:,:,:,m,:);
    sz(d) = size(x,d);
    x = reshape(x,sz);

else
    % slightly slower, but very unlikely to happen in practice.

    nd = ndims(x);
    sz = [sz ones(1,max(0,d-nd))];
    x = reshape(x,[prod(sz(1:(d-1))) sz(d) prod(sz((d+1):end))]);
    x = x(:,m,:);
    sz(d) = size(x,2); % be careful!
    x = reshape(x,sz);

end









function checkmask(m,len)

if ~islogical(m)
    error('m should be logical');
end

if length(m)~=len
    fprintf('length(mask)=%d size(.,dim)=%d\n',length(m),len);
    error('x,m dimensions are inconsistent');
end

if len > 0 && ~isvector(m)
    error('m should be vector');
end



