function Q = cn_sparsifypbdist(Q,shape,N,K,sparsityk,USEMEX)
% Approximate a discrete probability distribution with a fixed number of top probabilities.
% Especially useful for mixture model (eg, dictionary) posteriors.
% 
% Q         Probability distribution matrix. Each row (NxK) or column (KxN) sums to 1. (eg, posterior probabilities in MoG)
% shape     'NxK' | 'KxN'
% [N]       Number of data points.
% [K]       Number of categories.
% sparsityk Number of top entries to keep. If inf, result is dense.
%
% OUTPUT 
% Q         Re-normalized approximate probabilities, sparse matrix. 
%
% R.Gokberk Cinbis, Jakob Verbeek Nov 2011

% Note: This approximation can be quantified easily based on KL divergence.

if issparse(Q)
    mywarn('WARNING: cn_sparsifypbdist::Input Q is already sparse. (Will still process)');
end

q_tic = tic;

if ~exist('USEMEX','var')
    % Both methods seem equivalent and mex is much faster.
    USEMEX = true;
end

switch(shape)
    case 'NxK'

        if isempty(N)
            N = size(Q,1);
        end
        if isempty(K)
            K = size(Q,2);
        end

        assert(isequal(size(Q),[N K]));

        if ~isinf(sparsityk)

            if USEMEX
                [xs,j] = cn_sorttopk(Q,sparsityk,2,'descend');
            else
                % most time is spenty here
                [xs,j] = sort(Q,2,'descend'); 
                j=j(:,1:sparsityk);  % (N .)
                xs=xs(:,1:sparsityk);  % (N .)
            end
            xs=inorm(xs,2); % necessary, output should be a proper distribution
            i=repmat((1:N)',1,sparsityk);
            Q = sparse(i(:),j(:),xs(:),size(Q,1),size(Q,2)); % (N K) sparse.
        end

    case 'KxN'

        if isempty(N)
            N = size(Q,2);
        end
        if isempty(K)
            K = size(Q,1);
        end
        assert(isequal(size(Q),[K N]));

        if ~isinf(sparsityk)
            if USEMEX
                [xs,i] = cn_sorttopk(Q,sparsityk,1,'descend');
            else
                [xs,i] = sort(Q,1,'descend'); 
                i=i(1:sparsityk,:);  % (. N)
                xs=xs(1:sparsityk,:);  % (. N)
            end
            xs=inorm(xs,1); % necessary, output should be a proper distribution
            j=repmat(1:size(xs,2),sparsityk,1);
            Q = sparse(i(:),j(:),xs(:),size(Q,1),size(Q,2)); % (K N) sparse.
        end

    otherwise
        error('unknown');
end

%fprintf('mfa_sparsity: '); toc(q_tic);






function W = inorm(W,dim)

% Bug fix on 10-01-2012 [should not matter]:
% 1./5.2509e-309! So x(x==0)=1. is not a good solution.

x = sum(W,dim);
x(x<realmin)=realmin;
W = bsxfun(@times,W,1./x);





