
function d = prepare_descriptors(cache_level)
  % Channels = components of the full Fisher descriptor
  % loads PQ information + descriptors for each channel

  % servers used for caching if cache_level == 2 (otherwise all data is local)
    
  servers(1) = struct('name', 'localhost', ... % hostname the server runs on
		      'port', 1234, ...        % port 
		      'last_image', 5);        % serves images up to this index
  servers(2) = struct('name', 'localhost', 'port', 1235, 'last_image', 10); 

  
  pqs = {};
  channels = {};
  z2c = []; % maps channel no to PQ no  
  dirnames = {};
  for desc = {'sift', 'color'}
    desc = char(desc);
    
    % SPM grid
    if isequal(desc, 'sift')
      % grids =  {[1 1],[4 4]}; % for full dataset
      grids =  {[1 1], [2 1]}; % for toy dataset
    else
      grids = {[1 1]};
    end
    
    for src = {'boxes', 'masks'}
      src = char(src);
      pq_file = [ 'data/trained/PQ_centroids_' desc '_for_' src '.mat'];
      load(pq_file);
      
      % update centroids with power-law
      for i = 1:length(pq.centroids)
        pq.centroids{i} = sign(pq.centroids{i}) .* sqrt(abs(pq.centroids{i}));
      end
      
      % make 3D table of centroids
      pq = pq_prep(pq);
      
      % precompute per-centrois L2 norms 
      pq.l2normsq = reshape(sum(double(pq.cent3s).^2,1),256,pq.nsq);

      pqs{end + 1} = pq;
      
      for grid = grids
        
        idim = grid{1}(1);
        jdim = grid{1}(2);
        
        % order of i,j loops matters
        for jgrid = 1:jdim
          for igrid = 1:idim
            
            name = sprintf('%s_%s_%dx%di%dj%d', desc, src, idim, jdim, igrid, jgrid);
            
            
            root_dir = ['data/fisher_vectors/' name ];
            
            channel = struct('name', name, 'isfullim', false, 'ismask', isequal(src, 'masks'), 'root_dir', root_dir); 
            
            if cache_level == 0            
              channel.gt = load_flexdata(root_dir, 'gt');
              channel.bb = load_flexdata(root_dir, 'bb');
            end
            
            channels{end + 1} = channel;
            z2c(length(channels)) = length(pqs);
          end
        end
      end       

      if isequal(src, 'boxes')
        % full-image descriptor only in non-masked version
        name = sprintf('%s_%s_1x1i1j1', desc, src);
        
        root_dir = ['data/fisher_vectors/' name ];
        
        channel = struct('name', name, 'isfullim', true, 'ismask', false, 'pq', pq, 'root_dir', root_dir);
        
        channels{end + 1} = channel;
        z2c(length(channels)) = length(pqs);
      end
            
    end      
        
  end
  
  d.channels = {channels{end:-1:1}}; % for sync with Gokberk's code
  d.z2c = z2c(end:-1:1);
  d.pqs = [pqs{:}];
  
  if cache_level == 0
    fprintf('Loading all database descriptors\n');
    for i = 1:length(d.channels)
      fprintf('  Channel %s fullim: %d\n', d.channels{i}.name, d.channels{i}.isfullim);
      if d.channels{i}.isfullim
        d.channels{i}.im = load_flexdata(d.channels{i}.root_dir, 'im');
      else         
        d.channels{i}.gt = load_flexdata(d.channels{i}.root_dir, 'gt');
        d.channels{i}.bb = load_flexdata(d.channels{i}.root_dir, 'bb');        
      end      
    end    
  else
    fprintf('Loading in mex\n');
    imagemask = []; 
    basedirs = {}; 
    for i = 1:length(d.channels)
      basedirs{i} = d.channels{i}.root_dir; 
      imagemask(i) = d.channels{i}.isfullim; 
    end    
    if cache_level == 1
      descriptor_cache_mex('load', basedirs, int32(imagemask), true);
    else       
      descriptor_cache_mex('load-servers', basedirs, int32(imagemask), servers);
    end
    
  end
  d.cache_level = cache_level; 
  
end



function b = load_flexdata(dirname, fieldname) 
  fh(1) = fopen(sprintf('%s/npts_%s.info', dirname, fieldname), 'r');
  fh(2) = fopen(sprintf('%s/npts_%s.data', dirname, fieldname), 'r');
  fh(3) = fopen(sprintf('%s/pq_%s.info', dirname, fieldname), 'r');
  fh(4) = fopen(sprintf('%s/pq_%s.bloscinfo', dirname, fieldname), 'r');
  fh(5) = fopen(sprintf('%s/pq_%s.blosc', dirname, fieldname), 'r');
  assert(min(fh) > -1);
  
  % load headers
  header_npts = fread(fh(1), 5, 'uint64');
  header_codes = fread(fh(3), 5, 'uint64');
  N = fread(fh(4), 1, 'uint64');
  
  assert(all(header_npts(1:4) == [1 2 4 2]'));
  assert(all(header_codes(1:4) == [1 0 1 2]'));
  assert(header_npts(5) == header_codes(5) && header_codes(5) == N);

  % load info files
  npts_sizes = fread(fh(1), [2 N], 'uint64');
  code_sizes = fread(fh(3), [2 N], 'uint64');
  
  assert(all(npts_sizes(1, :) == 1));
  assert(all(code_sizes(1, :) == code_sizes(1, 1)));
  assert(all(npts_sizes(2, :) == code_sizes(2, :)));
  
  blosc_sizes = fread(fh(4), N, 'uint64');

  % load data
  b = {};
  for i = 1:N
    b{i}.npts = fread(fh(2), npts_sizes(:, i)', 'single');
    size = double(code_sizes(:, i));
    b{i}.codes = decompress_blosc_mex(fread(fh(5), blosc_sizes(i), 'uint8=>uint8'), size);
  end

  b = [b{:}];
  for i = 1:5
    fclose(fh(i));
  end
  
end
