function trees = cn_selectivesearchsegmentation(im)
% trees = cn_selectivesearchsegmentation(p)
%
% Return hierarchies that are used to generate candidate boxes in 
%   Segmentation as Selective Search for Object Recognition, ICCV 2011.
%
% In order to get the final set of boxes, use
%   unique(vertcat(trees.bbs),'rows')
% This usually removes ~20% of the boxes.
%
% INPUT
% p
%   .im
%
% OUTPUT
% trees(i)      Struct array of hierarchies.
%   .nblob      (scalar)    Total number of blobs in the hierarchy.
%   .nleaf      (scalar)    Number of leaf nodes in the hierarchy.
%   .leafmap    (imh imw)   [uint16 or uint32] segmentation map of leaves with ids 1..nleaf      
%   .parents    (1 nblob)   Denotes the parent of each node. First "nleaf" nodes are the leaf nodes. parents(end)=0 (ie, root).
%   .bbs        (nblob 4)   int16, each row is (x1 y1 x2 y2).
%
% SEE cn_plottree cn_showhiersegments
%
% R.G.Cinbis, Oct 2012

persistent initialize

DEBUG = false;

if isempty(initialize) % much faster.. not surprisingly.
    warning('off','images:removing:function');
    initialize = false;
end

% these options generate equivalent results to the default boxes published on author webpages.
colorTypes = {'Rgb', 'Hsv', 'RGI', 'Opp'};
kThresholds = [100 200]; 
sigma = 0.8;
numHierarchy = length(colorTypes) * length(kThresholds);
imsz = [size(im,1) size(im,2)];

trees(numHierarchy).nleaf = [];

idx = 1; % hierarchy id
for k = kThresholds
    minSize = k; % We use minSize = k.
    for colorTypeI = 1:length(colorTypes)
        colorType = colorTypes{colorTypeI};

        %currBox{idx} = SelectiveSearch(im, sigma, k, minSize, colorType);
        [blobs parents]= Image2HierarchicalGroupingWithOptions(im, sigma, k, minSize, {'Size','Texture'}, colorType);
        blobs = [blobs{:}];

        % cn_plottree(struct2('parents',parents,'showlabels',false));

        % -- gokberk --
        %
        % blobs(i)
        %   .rect: [y1 x1 y2 x2]
        %   .mask: [hxw logical], h=y2-y1+1, w=x2-x1+1
        % parents(i) is the parent of the blobs(i), and length(parents)=length(blobs)-1
        % blobs(end) is the parent of all blobs, therefore, it has no parent.
        %
        % find the leaf nodes, which are necessary to obtain upper level nodes.
        % if it has a child node, it is not a leaf node.
        %
        % hierarchy is a binary tree according to the paper. as far as I see, 
        % each node has either 0 (leaf level) or 2 children.
        %
        % logically and as far as I see, leaf level segments dont overlap.

        % find the number of leaf and make sure that they are only the first "nleaf" blobs 
        isleaf = true(1,length(blobs));
        isleaf(parents) = false;
        leafind = sort(find(isleaf));
        assert(leafind(end)==length(leafind));
        nleaf = length(detach(leafind));

        % convert leaf level into segmentation map
        if nleaf > 65535
            assert(nleaf <= 4294967295);
            leafmap = zeros(imsz,'uint32');
        else
            leafmap = zeros(imsz,'uint16');
        end
        for bi = 1:nleaf
            leafmap = setblob(leafmap,blobs,bi,DEBUG);
        end

        % set output
        trees(idx).nblob   = length(blobs);
        trees(idx).nleaf   = nleaf;
        trees(idx).leafmap = leafmap;
        trees(idx).parents = [parents 0];
        r = vertcat(blobs.rect);
        trees(idx).bbs     = int16( r(:,[2 1 4 3]) );

        if DEBUG
            if cn_timepassth; mywarn('cn_selectivesearchsegmentation:extra check! (2)'); end;
            assert(all(leafmap(:)),'leaf level does not cover the image!'); 

            % check that hierachy is compliant with the returned segments
            children = cn_treeparents2children([parents(:)' 0]);
            for bi = (nleaf+1):length(blobs)
                assert( isequal( setblob(false(imsz),blobs,bi,false), ismember(leafmap,getleafchildren(children,bi) )));
            end
        end

        idx = idx + 1;
    end
end







function leafmap = setblob(leafmap,blobs,bi,DEBUG)
% update leafmap to contain

r = blobs(bi).rect;
mask = blobs(bi).mask;
x = leafmap(r(1):r(3),r(2):r(4)); 

if DEBUG
    if cn_timepassth; mywarn('cn_selectivesearchsegmentation:extra check!'); end;
    if any(colvec(x(mask))); error('leaflevel overlap'); end; 
end

x(mask) = bi;
leafmap(r(1):r(3),r(2):r(4)) = x;







function leafs = getleafchildren(children,nodei)
% find all leaf nodes recusively making up a particular node.
% result is a vector of indices.

xc = children{nodei};
n = length(xc);
if n==0
    leafs = nodei;
else
    leafs = cell(1,n);
    for j = 1:n
        leafs{j} = getleafchildren(children,xc(j));
    end
    leafs = [leafs{:}];
end

    


